#!/usr/bin/env python3
# _*_ coding: utf-8 _*_

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk
from azote.tools import create_pixbuf
from azote.color_tools import hex_to_rgb
from azote import common

# Check if python yaml module available
try:
    from yaml import load, dump
    common.env['yaml'] = True
except Exception as e:
    common.env['yaml'] = False

if common.env['yaml']:
    try:
        from yaml import CLoader as Loader, CDumper as Dumper
    except ImportError:
        from yaml import Loader, Dumper


class Alacritty(Gtk.Window):
    def __init__(self):
        super().__init__()

        self.set_title('alacritty.yml')
        self.set_resizable(False)
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.set_position(Gtk.WindowPosition.NONE)
        self.set_keep_above(True)

        vbox0 = Gtk.VBox()
        vbox0.set_spacing(5)
        vbox0.set_border_width(5)
        
        hbox0 = Gtk.HBox()
        hbox0.set_spacing(5)
        hbox0.set_border_width(5)
        
        f = open(common.alacritty_config, "rb")
        self.data = load(f, Loader=Loader)
        try:
            output = dump(self.data['colors'], Dumper=Dumper, default_flow_style=False, sort_keys=False)
        except KeyError:
            output = None

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_propagate_natural_width(True)

        self.textview = Gtk.TextView()
        self.textview.set_property("name", "preview")
        self.textview.set_editable(False)

        self.textbuffer = self.textview.get_buffer()
        if output:
            self.textbuffer.set_text(output)
        else:
            self.textbuffer.set_text("No color definitions found")
        scrolled_window.add(self.textview)
        
        if output:
            hbox0.add(scrolled_window)
    
            vbox = Gtk.VBox()
            vbox.set_spacing(3)
            vbox.set_border_width(5)
    
            if self.data['colors']:
                for key in self.data['colors']:
                    label = Gtk.Label()
                    label.set_property("name", "dotfiles-header")
                    label.set_text(key.upper())
                    vbox.add(label)
                    try:
                        for key1 in self.data['colors'][key]:
                            hbox = Gtk.HBox()
                            label = Gtk.Label()
                            label.set_property("name", "dotfiles")
                            label.set_text(key1)
                            hbox.pack_start(label, True, False, 0)
                            label = Gtk.Label()
                            label.set_property("name", "dotfiles")
                            hex_color = self.data['colors'][key][key1].replace('0x', '#')
                            label.set_text(hex_color)
                            hbox.pack_start(label, True, False, 0)

                            preview_box = ColorPreviewBox(hex_color)
                            preview_box.connect('button-press-event', self.on_box_press, label, key, key1)

                            hbox.pack_start(preview_box, False, False, 0)

                            vbox.pack_start(hbox, False, False, 0)
                    except:
                        pass
    
            hbox0.add(vbox)
    
            vbox0.add(hbox0)

        hbox = Gtk.HBox()
        hbox.set_spacing(5)
        hbox.set_border_width(5)
        label = Gtk.Label()
        if output:
            label.set_text(common.lang['copy_paste_into'].format(common.alacritty_config))
        else:
            label.set_text(common.lang['no_colour_definitions'].format(common.alacritty_config))
        label.set_property('name', 'dotfiles')
        hbox.add(label)
        button = Gtk.Button.new_with_label(common.lang['close'])
        button.connect_after('clicked', self.close_window)
        hbox.pack_start(button, False, False, 0)

        vbox0.pack_start(hbox, False, False, 0)

        self.add(vbox0)
        self.show_all()

    def update_preview(self):
        output = dump(self.data['colors'], Dumper=Dumper, default_flow_style=False, sort_keys=False)
        self.textbuffer.set_text(output)
        
    def on_box_press(self, preview_box, event, label, section, key):
        if common.clipboard_text:
            self.data['colors'][section][key] = common.clipboard_text.replace('#', '0x')
            label.set_text(common.clipboard_text)
            preview_box.update()
            self.update_preview()
            
    def close_window(self, button):
        self.close()


class Xresources(Gtk.Window):
    def __init__(self):
        super().__init__()

        self.set_title('.Xresources')
        self.set_resizable(False)
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.set_position(Gtk.WindowPosition.NONE)
        self.set_keep_above(True)

        vbox0 = Gtk.VBox()
        vbox0.set_spacing(5)
        vbox0.set_border_width(5)

        hbox0 = Gtk.HBox()
        hbox0.set_spacing(5)
        hbox0.set_border_width(5)

        f = open(common.xresources, "r")
        lines = f.read().splitlines()
        f.close()
        self.data = {}

        # We only parse the lines 2 or 3 words long; the last one must be a hex color like '#rrggbb'.
        for line in lines:
            line = line.strip()
            parts = line.split()

            if 0 < len(parts) < 4 and parts[-1].startswith('#') and len(parts[-1]) == 7:
                try:
                    rgb = hex_to_rgb(parts[-1])  # validate the hex colour value
                    if len(parts) == 2:
                        key, value = parts
                        self.data[key] = value

                    elif len(parts) == 3:
                        keyword, name, value = parts
                        key = '{} {}'.format(keyword, name)
                        self.data[key] = value
                except ValueError:
                    print('Improper color value', parts[-1])

        output = ''
        for key, value in self.data.items():
            output += '{}  {}\n'.format(key, value)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_propagate_natural_width(True)

        self.textview = Gtk.TextView()
        self.textview.set_property("name", "preview")
        self.textview.set_editable(False)

        self.textbuffer = self.textview.get_buffer()
        self.textbuffer.set_text(output)
        scrolled_window.add(self.textview)

        hbox0.add(scrolled_window)

        vbox = Gtk.VBox()
        vbox.set_spacing(3)
        vbox.set_border_width(5)

        for key, value in self.data.items():
            hbox = Gtk.HBox()
            label = Gtk.Label()
            label.set_property("name", "dotfiles")
            label.set_text(key)
            hbox.pack_start(label, True, False, 0)
            label = Gtk.Label()
            label.set_property("name", "dotfiles")
            hex_color = self.data[key]
            label.set_text(hex_color)
            hbox.pack_start(label, True, False, 0)

            preview_box = ColorPreviewBox(hex_color)
            preview_box.connect('button-press-event', self.on_box_press, label, key)

            hbox.pack_start(preview_box, False, False, 0)

            vbox.pack_start(hbox, False, False, 0)

        hbox0.add(vbox)

        vbox0.add(hbox0)

        hbox = Gtk.HBox()
        hbox.set_spacing(5)
        hbox.set_border_width(5)
        label = Gtk.Label(common.lang['copy_paste_into'].format(common.xresources))
        label.set_property('name', 'dotfiles')
        hbox.add(label)
        button = Gtk.Button.new_with_label(common.lang['close'])
        button.connect_after('clicked', self.close_window)
        hbox.pack_start(button, False, False, 0)

        vbox0.pack_start(hbox, False, False, 0)

        self.add(vbox0)
        self.show_all()

    def update_preview(self):
        output = ''
        for key, value in self.data.items():
            output += '{}  {}\n'.format(key, value)
        self.textbuffer.set_text(output)

    def on_box_press(self, preview_box, event, label, key):
        if common.clipboard_text:
            self.data[key] = common.clipboard_text
            label.set_text(common.clipboard_text)
            preview_box.update()
            self.update_preview()

    def close_window(self, button):
        self.close()


class ColorPreviewBox(Gtk.EventBox):
    def __init__(self, hex_color):
        super().__init__()
        try:
            pixbuf = create_pixbuf((common.settings.clip_prev_size, common.settings.clip_prev_size // 2),
                               hex_to_rgb(hex_color))
        except:
            print('Improper color value: {}'.format(hex_color))
            pixbuf = create_pixbuf((common.settings.clip_prev_size, common.settings.clip_prev_size // 2),
                                   hex_to_rgb('#000000'))
        self.gtk_image = Gtk.Image.new_from_pixbuf(pixbuf)
        self.add(self.gtk_image)
        
    def update(self):
        if common.clipboard_text:
            pixbuf = create_pixbuf((common.settings.clip_prev_size, common.settings.clip_prev_size // 2),
                                   hex_to_rgb(common.clipboard_text))
            self.gtk_image.set_from_pixbuf(pixbuf)



#!/usr/bin/env python3
# _*_ coding: utf-8 _*_

"""
Wallpaper manager for Sway, i3 and some other WMs, as a frontend to swaybg and feh

Author: Piotr Miller
e-mail: nwg.piotr@gmail.com
Website: http://nwg.pl
Project: https://github.com/nwg-piotr/azote
License: GPL3
"""
app_name = 'azote'

CRITICAL = 'critical'
ERROR = 'error'
WARNING = 'warning'
INFO = 'info'
DEBUG = 'debug'

env = {}
sway = False
screen_h = None

lang = None             # dictionary "name": lang_string

preview = None
progress_bar = None
status_bar = None
thumbnails_list = None
display_boxes_list = None
selected_wallpaper = None
selected_picture_label = None
split_button = None
apply_button = None
apply_to_all_button = None

cols = 3                # number of columns in pictures preview

allowed_file_types = ['jpg', 'jpeg', 'png', 'webp']
associations = None     # dictionary {'extension": [program1, program2, program3, ...]}

app_dir = ''            # ~/.azote
thumb_dir = ''          # ~/.azote/thumbnails
tmp_dir = ''            # ~/.azote/temp
bcg_dir = ''            # ~/.azote/backgrounds-sway or ~/.azote/backgrounds-feh
sample_dir = ''         # ~/.azote/sample
log_file = ''           # ~/.azote/log.txt
cmd_file = ''           # ~/.azote/command.sh
config_home = ''
azote_config_home = ''        # $XDG_CONFIG_HOME or ~/.config/azote
data_home = ''          # $XDG_DATA_HOME or ~/.local/share/azote
alacritty_config = ''
xresources = ''

data_migrated = False

logging_enabled = True
displays = None         # detected displays details

settings = None         # object saved to / restored from ~/.azote/settings.pkl

modes_swaybg = ["stretch", "fit", "fill", "center", "tile"]
modes_feh = ["scale", "max", "fill", "center", "tile"]

main_window = None
clipboard = None
clipboard_text = ''     # to transfer colors between toolbars we'll use this instead of the real clipboard content
picker = False

cpd = None              # ColorPaletteDialog object
dotfile_window = None
picker_window = None
indicator = None

color_names = None



#!/usr/bin/env python3
# _*_ coding: utf-8 _*_


def rgba_to_hex(color):
    """
    Return hexadecimal string for :class:`Gdk.RGBA` `color`
    http://wrhansen.blogspot.com/2012/09/how-to-convert-gdkrgba-to-hex-string-in.html
    """
    return "#{0:02x}{1:02x}{2:02x}".format(int(color.red * 255),
                                           int(color.green * 255),
                                           int(color.blue * 255))


def rgb_to_hex(rgb_color):
    return '#%02x%02x%02x' % (rgb_color[0], rgb_color[1], rgb_color[2])


def hex_to_rgb(string):
    string = string.lstrip('#')
    return tuple(int(string[i:i+2], 16) for i in (0, 2, 4))


def rgb_to_rgba(rgb):
    """
    :param rgb: tuple (rrr, ggg, bbb)
    :return: tuple (r.r, g.g, b.b, 1.0)
    """
    rgba = (rgb[0] / 255, rgb[1] / 255, rgb [2] / 255, 1.0)
    return rgba


class WikiColours(object):
    def __init__(self):
        super().__init__()
        # Definitions come from https://en.wikipedia.org/wiki/List_of_colors:_A%E2%80%93F.
        # I removed redundant names arbitrarily.
        self.colours = {"#0048ba": "Absolute Zero",
                        "#b0bf1a": "Acid green",
                        "#7cb9e8": "Aero",
                        "#c9ffe5": "Aero blue",
                        "#b284be": "African violet",
                        "#72a0c1": "Air superiority blue",
                        "#edeae0": "Alabaster",
                        "#f0f8ff": "Alice blue",
                        "#c46210": "Alloy orange",
                        "#efdecd": "Almond",
                        "#e52b50": "Amaranth",
                        "#9f2b68": "Amaranth (M&P)",
                        "#f19cbb": "Amaranth pink",
                        "#ab274f": "Amaranth purple",
                        "#d3212d": "Amaranth red",
                        "#3b7a57": "Amazon",
                        "#ffbf00": "Amber",
                        "#ff7e00": "Amber (SAE/ECE)",
                        "#9966cc": "Amethyst",
                        "#a4c639": "Android green",
                        "#cd9575": "Antique brass",
                        "#665d1e": "Antique bronze",
                        "#915c83": "Antique fuchsia",
                        "#841b2d": "Antique ruby",
                        "#faebd7": "Antique white",
                        "#8db600": "Apple green",
                        "#fbceb1": "Apricot",
                        "#7fffd4": "Aquamarine",
                        "#d0ff14": "Arctic lime",
                        "#4b5320": "Army green",
                        "#8f9779": "Artichoke",
                        "#e9d66b": "Arylide yellow",
                        "#b2beb5": "Ash gray",
                        "#87a96b": "Asparagus",
                        "#ff9966": "Atomic tangerine",
                        "#a52a2a": "Auburn",
                        "#fdee00": "Aureolin",
                        "#568203": "Avocado",
                        "#007fff": "Azure",
                        "#f0ffff": "Azure (X11/web color)",
                        "#89cff0": "Baby blue",
                        "#a1caf1": "Baby blue eyes",
                        "#f4c2c2": "Baby pink",
                        "#fefefa": "Baby powder",
                        "#ff91af": "Baker-Miller pink",
                        "#fae7b5": "Banana Mania",
                        "#da1884": "Barbie Pink",
                        "#7c0a02": "Barn red",
                        "#bcd4e6": "Beau blue",
                        "#9f8170": "Beaver",
                        "#f5f5dc": "Beige",
                        "#2e5894": "B'dazzled blue",
                        "#9c2542": "Big dip o’ruby",
                        "#ffe4c4": "Bisque",
                        "#3d2b1f": "Bistre",
                        "#967117": "Bistre brown",
                        "#cae00d": "Bitter lemon",
                        "#bfff00": "Bitter lime",
                        "#fe6f5e": "Bittersweet",
                        "#bf4f51": "Bittersweet shimmer",
                        "#000000": "Black",
                        "#3d0c02": "Black bean",
                        "#1b1811": "Black chocolate",
                        "#3b2f2f": "Black coffee",
                        "#54626f": "Black coral",
                        "#3b3c36": "Black olive",
                        "#bfafb2": "Black Shadows",
                        "#ffebcd": "Blanched almond",
                        "#a57164": "Blast-off bronze",
                        "#318ce7": "Bleu de France",
                        "#ace5ee": "Blizzard blue",
                        "#faf0be": "Blond",
                        "#660000": "Blood red",
                        "#0000ff": "Blue",
                        "#1f75fe": "Blue (Crayola)",
                        "#0093af": "Blue (Munsell)",
                        "#0087bd": "Blue (NCS)",
                        "#0018a8": "Blue (Pantone)",
                        "#333399": "Blue (pigment)",
                        "#0247fe": "Blue (RYB)",
                        "#a2a2d0": "Blue bell",
                        "#6699cc": "Blue-gray",
                        "#0d98ba": "Blue-green",
                        "#064e40": "Blue-green (color wheel)",
                        "#5dadec": "Blue jeans",
                        "#126180": "Blue sapphire",
                        "#8a2be2": "Blue-violet",
                        "#7366bd": "Blue-violet (Crayola)",
                        "#4d1a7f": "Blue-violet (color wheel)",
                        "#5072a7": "Blue yonder",
                        "#3c69e7": "Bluetiful",
                        "#de5d83": "Blush",
                        "#79443b": "Bole",
                        "#e3dac9": "Bone",
                        "#006a4e": "Bottle green",
                        "#87413f": "Brandy",
                        "#cb4154": "Brick red",
                        "#66ff00": "Bright green",
                        "#d891ef": "Bright lilac",
                        "#c32148": "Bright maroon",
                        "#1974d2": "Bright navy blue",
                        "#ffaa1d": "Bright yellow (Crayola)",
                        "#ff55a3": "Brilliant rose",
                        "#fb607f": "Brink pink",
                        "#004225": "British racing green",
                        "#cd7f32": "Bronze",
                        "#88540b": "Brown",
                        "#af6e4d": "Brown sugar",
                        "#1b4d3e": "Brunswick green",
                        "#7bb661": "Bud green",
                        "#f0dc82": "Buff",
                        "#800020": "Burgundy",
                        "#deb887": "Burlywood",
                        "#a17a74": "Burnished brown",
                        "#cc5500": "Burnt orange",
                        "#e97451": "Burnt sienna",
                        "#8a3324": "Burnt umber",
                        "#bd33a4": "Byzantine",
                        "#702963": "Byzantium",
                        "#536872": "Cadet",
                        "#5f9ea0": "Cadet blue",
                        "#a9b2c3": "Cadet blue (Crayola)",
                        "#91a3b0": "Cadet grey",
                        "#006b3c": "Cadmium green",
                        "#ed872d": "Cadmium orange",
                        "#e30022": "Cadmium red",
                        "#fff600": "Cadmium yellow",
                        "#a67b5b": "Café au lait",
                        "#4b3621": "Café noir",
                        "#a3c1ad": "Cambridge blue",
                        "#efbbcc": "Cameo pink",
                        "#ffff99": "Canary",
                        "#ffef00": "Canary yellow",
                        "#ff0800": "Candy apple red",
                        "#e4717a": "Candy pink",
                        "#592720": "Caput mortuum",
                        "#c41e3a": "Cardinal",
                        "#00cc99": "Caribbean green",
                        "#960018": "Carmine",
                        "#d70040": "Carmine (M&P)",
                        "#ffa6c9": "Carnation pink",
                        "#b31b1b": "Carnelian",
                        "#56a0d3": "Carolina blue",
                        "#ed9121": "Carrot orange",
                        "#00563f": "Castleton green",
                        "#703642": "Catawba",
                        "#c95a49": "Cedar Chest",
                        "#ace1af": "Celadon",
                        "#007ba7": "Celadon blue",
                        "#2f847c": "Celadon green",
                        "#246bce": "Celtic blue",
                        "#de3163": "Cerise",
                        "#2a52be": "Cerulean blue",
                        "#6d9bc3": "Cerulean frost",
                        "#1dacd6": "Cerulean (Crayola)",
                        "#007aa5": "CG blue",
                        "#e03c31": "CG red",
                        "#f7e7ce": "Champagne",
                        "#f1ddcf": "Champagne pink",
                        "#36454f": "Charcoal",
                        "#232b2b": "Charleston green",
                        "#e68fac": "Charm pink",
                        "#dfff00": "Chartreuse (traditional)",
                        "#7fff00": "Chartreuse (web)",
                        "#ffb7c5": "Cherry blossom pink",
                        "#954535": "Chestnut",
                        "#a8516e": "China rose",
                        "#aa381e": "Chinese red",
                        "#856088": "Chinese violet",
                        "#ffb200": "Chinese yellow",
                        "#7b3f00": "Chocolate (traditional)",
                        "#d2691e": "Chocolate (web)",
                        "#ffa700": "Chrome yellow",
                        "#98817b": "Cinereous",
                        "#e34234": "Cinnabar",
                        "#cd607e": "Cinnamon Satin",
                        "#e4d00a": "Citrine",
                        "#9fa91f": "Citron",
                        "#7f1734": "Claret",
                        "#0047ab": "Cobalt blue",
                        "#6f4e37": "Coffee",
                        "#b9d9eb": "Columbia Blue",
                        "#8c92ac": "Cool grey",
                        "#b87333": "Copper",
                        "#da8a67": "Copper (Crayola)",
                        "#ad6f69": "Copper penny",
                        "#cb6d51": "Copper red",
                        "#996666": "Copper rose",
                        "#ff3800": "Coquelicot",
                        "#ff7f50": "Coral",
                        "#893f45": "Cordovan",
                        "#6495ed": "Cornflower blue",
                        "#fff8dc": "Cornsilk",
                        "#2e2d88": "Cosmic cobalt",
                        "#fff8e7": "Cosmic latte",
                        "#81613c": "Coyote brown",
                        "#ffbcd9": "Cotton candy",
                        "#fffdd0": "Cream",
                        "#dc143c": "Crimson",
                        "#9e1b32": "Crimson (UA)",
                        "#f5f5f5": "Cultured",
                        "#00ffff": "Cyan",
                        "#00b7eb": "Cyan (process)",
                        "#58427c": "Cyber grape",
                        "#f56fa1": "Cyclamen",
                        "#666699": "Dark blue-gray",
                        "#654321": "Dark brown",
                        "#5d3954": "Dark byzantium",
                        "#26428b": "Dark cornflower blue",
                        "#008b8b": "Dark cyan",
                        "#536878": "Dark electric blue",
                        "#b8860b": "Dark goldenrod",
                        "#013220": "Dark green",
                        "#006400": "Dark green (X11)",
                        "#1a2421": "Dark jungle green",
                        "#bdb76b": "Dark khaki",
                        "#534b4f": "Dark liver",
                        "#543d37": "Dark liver (horses)",
                        "#8b008b": "Dark magenta",
                        "#4a5d23": "Dark moss green",
                        "#556b2f": "Dark olive green",
                        "#ff8c00": "Dark orange",
                        "#9932cc": "Dark orchid",
                        "#03c03c": "Dark pastel green",
                        "#301934": "Dark purple",
                        "#8b0000": "Dark red",
                        "#e9967a": "Dark salmon",
                        "#8fbc8f": "Dark sea green",
                        "#3c1414": "Dark sienna",
                        "#8cbed6": "Dark sky blue",
                        "#483d8b": "Dark slate blue",
                        "#2f4f4f": "Dark slate gray",
                        "#177245": "Dark spring green",
                        "#00ced1": "Dark turquoise",
                        "#9400d3": "Dark violet",
                        "#00703c": "Dartmouth green",
                        "#555555": "Davy's grey",
                        "#da3287": "Deep cerise",
                        "#b94e48": "Deep chestnut",
                        "#004b49": "Deep jungle green",
                        "#ff1493": "Deep pink",
                        "#ff9933": "Deep saffron",
                        "#00bfff": "Deep sky blue",
                        "#4a646c": "Deep Space Sparkle",
                        "#7e5e60": "Deep taupe",
                        "#1560bd": "Denim",
                        "#2243b6": "Denim blue",
                        "#edc9af": "Desert sand",
                        "#696969": "Dim gray",
                        "#1e90ff": "Dodger blue",
                        "#d71868": "Dogwood rose",
                        "#00009c": "Duke blue",
                        "#efdfbb": "Dutch white",
                        "#e1a95f": "Earth yellow",
                        "#555d50": "Ebony",
                        "#c2b280": "Ecru",
                        "#1b1b1b": "Eerie black",
                        "#614051": "Eggplant",
                        "#f0ead6": "Eggshell",
                        "#1034a6": "Egyptian blue",
                        "#7df9ff": "Electric blue",
                        "#6f00ff": "Electric indigo",
                        "#ccff00": "Electric lime",
                        "#bf00ff": "Electric purple",
                        "#6c3082": "Eminence",
                        "#b48395": "English lavender",
                        "#ab4b52": "English red",
                        "#cc474b": "English vermillion",
                        "#563c5c": "English violet",
                        "#00ff40": "Erin",
                        "#96c8a2": "Eton blue",
                        "#801818": "Falu red",
                        "#b53389": "Fandango",
                        "#de5285": "Fandango pink",
                        "#f400a1": "Fashion fuchsia",
                        "#e5aa70": "Fawn",
                        "#4d5d53": "Feldgrau",
                        "#4f7942": "Fern green",
                        "#6c541e": "Field drab",
                        "#ff5470": "Fiery rose",
                        "#b22222": "Firebrick",
                        "#ce2029": "Fire engine red",
                        "#e95c4b": "Fire opal",
                        "#e25822": "Flame",
                        "#eedc82": "Flax",
                        "#0063dc": "Flickr Blue",
                        "#fb0081": "Flickr Pink",
                        "#a2006d": "Flirt",
                        "#fffaf0": "Floral white",
                        "#15f4ee": "Fluorescent blue",
                        "#5fa777": "Forest green (Crayola)",
                        "#014421": "Forest green (traditional)",
                        "#228b22": "Forest green (web)",
                        "#856d4d": "French bistre",
                        "#0072bb": "French blue",
                        "#fd3f92": "French fuchsia",
                        "#86608e": "French lilac",
                        "#9efd38": "French lime",
                        "#d473d4": "French mauve",
                        "#fd6c9e": "French pink",
                        "#c72c48": "French raspberry",
                        "#f64a8a": "French rose",
                        "#77b5fe": "French sky blue",
                        "#8806ce": "French violet",
                        "#e936a7": "Frostbite",
                        "#c154c1": "Fuchsia (Crayola)",
                        "#cc397b": "Fuchsia purple",
                        "#c74375": "Fuchsia rose",
                        "#e48400": "Fulvous",
                        "#87421f": "Fuzzy Wuzzy",
                        "#dcdcdc": "Gainsboro",
                        "#e49b0f": "Gamboge",
                        "#007f66": "Generic viridian",
                        "#f8f8ff": "Ghost white",
                        "#6082b6": "Glaucous",
                        "#ab92b3": "Glossy grape",
                        "#00ab66": "GO green",
                        "#a57c00": "Gold",
                        "#d4af37": "Gold (metallic)",
                        "#ffd700": "Gold (web) (Golden)",
                        "#e6be8a": "Gold (Crayola)",
                        "#85754e": "Gold Fusion",
                        "#996515": "Golden brown",
                        "#fcc200": "Golden poppy",
                        "#ffdf00": "Golden yellow",
                        "#daa520": "Goldenrod",
                        "#676767": "Granite gray",
                        "#a8e4a0": "Granny Smith apple",
                        "#808080": "Gray (web)",
                        "#bebebe": "Gray (X11 gray)",
                        "#00ff00": "Green",
                        "#1cac78": "Green (Crayola)",
                        "#008000": "Green (web)",
                        "#00a877": "Green (Munsell)",
                        "#009f6b": "Green (NCS)",
                        "#00ad43": "Green (Pantone)",
                        "#00a550": "Green (pigment)",
                        "#66b032": "Green (RYB)",
                        "#1164b4": "Green-blue",
                        "#2887c8": "Green-blue (Crayola)",
                        "#009966": "Green-cyan",
                        "#a7f432": "Green Lizard",
                        "#6eaea1": "Green Sheen",
                        "#adff2f": "Green-yellow",
                        "#f0e891": "Green-yellow (Crayola)",
                        "#a99a86": "Grullo",
                        "#2a3439": "Gunmetal",
                        "#446ccf": "Han blue",
                        "#5218fa": "Han purple",
                        "#3fff00": "Harlequin",
                        "#da9100": "Harvest gold",
                        "#ff7a00": "Heat Wave",
                        "#df73ff": "Heliotrope",
                        "#aa98a9": "Heliotrope gray",
                        "#f0fff0": "Honeydew",
                        "#006db0": "Honolulu blue",
                        "#49796b": "Hooker's green",
                        "#ff1dce": "Hot magenta",
                        "#ff69b4": "Hot pink",
                        "#355e3b": "Hunter green",
                        "#71a6d2": "Iceberg",
                        "#fcf75e": "Icterine",
                        "#319177": "Illuminating emerald",
                        "#b2ec5d": "Inchworm",
                        "#4c516d": "Independence",
                        "#138808": "India green",
                        "#cd5c5c": "Indian red",
                        "#e3a857": "Indian yellow",
                        "#4b0082": "Indigo",
                        "#00416a": "Indigo dye",
                        "#002fa7": "International Klein Blue",
                        "#ff4f00": "International orange (aerospace)",
                        "#ba160c": "International orange (engineering)",
                        "#c0362c": "International orange (Golden Gate Bridge)",
                        "#5a4fcf": "Iris",
                        "#f4f0ec": "Isabelline",
                        "#b2ffff": "Italian sky blue",
                        "#fffff0": "Ivory",
                        "#00a86b": "Jade",
                        "#a50b5e": "Jazzberry jam",
                        "#343434": "Jet",
                        "#f4ca16": "Jonquil",
                        "#bdda57": "June bud",
                        "#29ab87": "Jungle green",
                        "#4cbb17": "Kelly green",
                        "#3ab09e": "Keppel",
                        "#e8f48c": "Key lime",
                        "#c3b091": "Khaki (web)",
                        "#f0e68c": "Khaki (X11) (Light khaki)",
                        "#e79fc4": "Kobi",
                        "#6b4423": "Kobicha",
                        "#354230": "Kombu green",
                        "#512888": "KSU purple",
                        "#d6cadd": "Languid lavender",
                        "#26619c": "Lapis lazuli",
                        "#ffff66": "Laser lemon",
                        "#a9ba9d": "Laurel green",
                        "#cf1020": "Lava",
                        "#b57edc": "Lavender (floral)",
                        "#e6e6fa": "Lavender (web)",
                        "#ccccff": "Lavender blue",
                        "#fff0f5": "Lavender blush",
                        "#c4c3d0": "Lavender gray",
                        "#7cfc00": "Lawn green",
                        "#fff700": "Lemon",
                        "#fffacd": "Lemon chiffon",
                        "#cca01d": "Lemon curry",
                        "#fdff00": "Lemon glacier",
                        "#f6eabe": "Lemon meringue",
                        "#fff44f": "Lemon yellow",
                        "#ffff9f": "Lemon yellow (Crayola)",
                        "#545aa7": "Liberty",
                        "#add8e6": "Light blue",
                        "#f08080": "Light coral",
                        "#93ccea": "Light cornflower blue",
                        "#e0ffff": "Light cyan",
                        "#c8ad7f": "Light French beige",
                        "#fafad2": "Light goldenrod yellow",
                        "#d3d3d3": "Light gray",
                        "#90ee90": "Light green",
                        "#fed8b1": "Light orange",
                        "#c5cbe1": "Light periwinkle",
                        "#ffb6c1": "Light pink",
                        "#ffa07a": "Light salmon",
                        "#20b2aa": "Light sea green",
                        "#87cefa": "Light sky blue",
                        "#778899": "Light slate gray",
                        "#b0c4de": "Light steel blue",
                        "#ffffe0": "Light yellow",
                        "#c8a2c8": "Lilac",
                        "#ae98aa": "Lilac Luster",
                        "#32cd32": "Lime green",
                        "#195905": "Lincoln green",
                        "#faf0e6": "Linen",
                        "#de6fa1": "Liseran purple",
                        "#6ca0dc": "Little boy blue",
                        "#674c47": "Liver",
                        "#b86d29": "Liver (dogs)",
                        "#6c2e1f": "Liver (organ)",
                        "#987456": "Liver chestnut",
                        "#ffbd88": "Macaroni and Cheese",
                        "#cc3336": "Madder Lake",
                        "#ff00ff": "Magenta",
                        "#f653a6": "Magenta (Crayola)",
                        "#ca1f7b": "Magenta (dye)",
                        "#d0417e": "Magenta (Pantone)",
                        "#ff0090": "Magenta (process)",
                        "#9f4576": "Magenta haze",
                        "#aaf0d1": "Magic mint",
                        "#f8f4ff": "Magnolia",
                        "#c04000": "Mahogany",
                        "#fbec5d": "Maize",
                        "#f2c649": "Maize (Crayola)",
                        "#6050dc": "Majorelle blue",
                        "#0bda51": "Malachite",
                        "#979aaa": "Manatee",
                        "#f37a48": "Mandarin",
                        "#fdbe02": "Mango",
                        "#ff8243": "Mango Tango",
                        "#74c365": "Mantis",
                        "#880085": "Mardi Gras",
                        "#eaa221": "Marigold",
                        "#800000": "Maroon (web)",
                        "#b03060": "Maroon (X11)",
                        "#e0b0ff": "Mauve",
                        "#915f6d": "Mauve taupe",
                        "#ef98aa": "Mauvelous",
                        "#47abcc": "Maximum blue",
                        "#30bfbf": "Maximum blue green",
                        "#acace6": "Maximum blue purple",
                        "#5e8c31": "Maximum green",
                        "#d9e650": "Maximum green yellow",
                        "#733380": "Maximum purple",
                        "#d92121": "Maximum red",
                        "#a63a79": "Maximum red purple",
                        "#fafa37": "Maximum yellow",
                        "#f2ba49": "Maximum yellow red",
                        "#4c9141": "May green",
                        "#73c2fb": "Maya blue",
                        "#66ddaa": "Medium aquamarine",
                        "#0000cd": "Medium blue",
                        "#e2062c": "Medium candy apple red",
                        "#af4035": "Medium carmine",
                        "#ba55d3": "Medium orchid",
                        "#9370db": "Medium purple",
                        "#3cb371": "Medium sea green",
                        "#7b68ee": "Medium slate blue",
                        "#00fa9a": "Medium spring green",
                        "#48d1cc": "Medium turquoise",
                        "#f8b878": "Mellow apricot",
                        "#f8de7e": "Mellow yellow",
                        "#febaad": "Melon",
                        "#d3af37": "Metallic gold",
                        "#0a7e8c": "Metallic Seaweed",
                        "#9c7c38": "Metallic Sunburst",
                        "#e4007c": "Mexican pink",
                        "#7ed4e6": "Middle blue",
                        "#8dd9cc": "Middle blue green",
                        "#8b72be": "Middle blue purple",
                        "#8b8680": "Middle grey",
                        "#4d8c57": "Middle green",
                        "#acbf60": "Middle green yellow",
                        "#d982b5": "Middle purple",
                        "#e58e73": "Middle red",
                        "#a55353": "Middle red purple",
                        "#ffeb00": "Middle yellow",
                        "#ecb176": "Middle yellow red",
                        "#702670": "Midnight",
                        "#191970": "Midnight blue",
                        "#004953": "Midnight green (eagle green)",
                        "#ffc40c": "Mikado yellow",
                        "#ffdae9": "Mimi pink",
                        "#e3f988": "Mindaro",
                        "#36747d": "Ming",
                        "#f5e050": "Minion yellow",
                        "#3eb489": "Mint",
                        "#f5fffa": "Mint cream",
                        "#98ff98": "Mint green",
                        "#bbb477": "Misty moss",
                        "#ffe4e1": "Misty rose",
                        "#8da399": "Morning blue",
                        "#8a9a5b": "Moss green",
                        "#30ba8f": "Mountain Meadow",
                        "#997a8d": "Mountbatten pink",
                        "#18453b": "MSU green",
                        "#c54b8c": "Mulberry",
                        "#c8509b": "Mulberry (Crayola)",
                        "#ffdb58": "Mustard",
                        "#317873": "Myrtle green",
                        "#d65282": "Mystic",
                        "#ad4379": "Mystic maroon",
                        "#f6adc6": "Nadeshiko pink",
                        "#ffdead": "Navajo white",
                        "#000080": "Navy blue",
                        "#4666ff": "Neon blue",
                        "#39ff14": "Neon green",
                        "#d7837f": "New York pink",
                        "#727472": "Nickel",
                        "#a4dded": "Non-photo blue",
                        "#e9ffdb": "Nyanza",
                        "#4f42b5": "Ocean Blue",
                        "#48bf91": "Ocean green",
                        "#cc7722": "Ochre",
                        "#43302e": "Old burgundy",
                        "#cfb53b": "Old gold",
                        "#fdf5e6": "Old lace",
                        "#796878": "Old lavender",
                        "#c08081": "Old rose",
                        "#848482": "Old silver",
                        "#808000": "Olive",
                        "#6b8e23": "Olive Drab (Nb3)",
                        "#3c341f": "Olive Drab (Nb7)",
                        "#b5b35c": "Olive green",
                        "#9ab973": "Olivine",
                        "#353839": "Onyx",
                        "#a8c3bc": "Opal",
                        "#b784a7": "Opera mauve",
                        "#ff7f00": "Orange",
                        "#ff7538": "Orange (Crayola)",
                        "#ff5800": "Orange (Pantone)",
                        "#ffa500": "Orange (web)",
                        "#ff9f00": "Orange peel",
                        "#ff681f": "Orange-red",
                        "#fa5b3d": "Orange soda",
                        "#f5bd1f": "Orange-yellow",
                        "#f8d568": "Orange-yellow (Crayola)",
                        "#da70d6": "Orchid",
                        "#f2bdcd": "Orchid pink",
                        "#e29cd2": "Orchid (Crayola)",
                        "#2d383a": "Outer space (Crayola)",
                        "#ff6e4a": "Outrageous Orange",
                        "#002147": "Oxford blue",
                        "#841617": "OU Crimson red",
                        "#1ca9c9": "Pacific blue",
                        "#006600": "Pakistan green",
                        "#682860": "Palatinate purple",
                        "#9bc4e2": "Pale cerulean",
                        "#fadadd": "Pale pink",
                        "#fae6fa": "Pale purple (Pantone)",
                        "#c9c0bb": "Pale silver",
                        "#78184a": "Pansy purple",
                        "#009b7d": "Paolo Veronese green",
                        "#ffefd5": "Papaya whip",
                        "#e63e62": "Paradise pink",
                        "#50c878": "Paris Green",
                        "#dea5a4": "Pastel pink",
                        "#ffe5b4": "Peach",
                        "#ffcba4": "Peach (Crayola)",
                        "#ffdab9": "Peach puff",
                        "#d1e231": "Pear",
                        "#b768a2": "Pearly purple",
                        "#c3cde6": "Periwinkle (Crayola)",
                        "#e12c2c": "Permanent Geranium Lake",
                        "#1c39bb": "Persian blue",
                        "#00a693": "Persian green",
                        "#32127a": "Persian indigo",
                        "#d99058": "Persian orange",
                        "#f77fbe": "Persian pink",
                        "#cc3333": "Persian red",
                        "#fe28a2": "Persian rose",
                        "#ec5800": "Persimmon",
                        "#8ba8b7": "Pewter Blue",
                        "#000f89": "Phthalo blue",
                        "#123524": "Phthalo green",
                        "#2e2787": "Picotee blue",
                        "#c30b4e": "Pictorial carmine",
                        "#fddde6": "Piggy pink",
                        "#01796f": "Pine green",
                        "#2a2f23": "Pine tree",
                        "#ffc0cb": "Pink",
                        "#d74894": "Pink (Pantone)",
                        "#fc74fd": "Pink flamingo",
                        "#ffddf4": "Pink lace",
                        "#d8b2d1": "Pink lavender",
                        "#f78fa7": "Pink Sherbet",
                        "#93c572": "Pistachio",
                        "#e5e4e2": "Platinum",
                        "#8e4585": "Plum",
                        "#dda0dd": "Plum (web)",
                        "#5946b2": "Plump Purple",
                        "#5da493": "Polished Pine",
                        "#be4f62": "Popstar",
                        "#ff5a36": "Portland Orange",
                        "#b0e0e6": "Powder blue",
                        "#f58025": "Princeton orange",
                        "#701c1c": "Prune",
                        "#003153": "Prussian blue",
                        "#df00ff": "Psychedelic purple",
                        "#cc8899": "Puce",
                        "#644117": "Pullman Brown (UPS Brown)",
                        "#ff7518": "Pumpkin",
                        "#6a0dad": "Purple",
                        "#800080": "Purple (web)",
                        "#9f00c5": "Purple (Munsell)",
                        "#a020f0": "Purple (X11)",
                        "#9678b6": "Purple mountain majesty",
                        "#4e5180": "Purple navy",
                        "#fe4eda": "Purple pizzazz",
                        "#9c51b6": "Purple Plum",
                        "#9a4eae": "Purpureus",
                        "#436b95": "Queen blue",
                        "#e8ccd7": "Queen pink",
                        "#a6a6a6": "Quick Silver",
                        "#8e3a59": "Quinacridone magenta",
                        "#ff355e": "Radical Red",
                        "#242124": "Raisin black",
                        "#fbab60": "Rajah",
                        "#e30b5d": "Raspberry",
                        "#b3446c": "Raspberry rose",
                        "#d68a59": "Raw Sienna",
                        "#826644": "Raw umber",
                        "#ff33cc": "Razzle dazzle rose",
                        "#e3256b": "Razzmatazz",
                        "#8d4e85": "Razzmic Berry",
                        "#663399": "Rebecca Purple",
                        "#ff0000": "Red",
                        "#ee204d": "Red (Crayola)",
                        "#f2003c": "Red (Munsell)",
                        "#c40233": "Red (NCS)",
                        "#ed2939": "Red (Pantone)",
                        "#ed1c24": "Red (pigment)",
                        "#fe2712": "Red (RYB)",
                        "#ff5349": "Red-orange",
                        "#ff4500": "Red-orange (Color wheel)",
                        "#e40078": "Red-purple",
                        "#fd3a4a": "Red Salsa",
                        "#c71585": "Red-violet",
                        "#c0448f": "Red-violet (Crayola)",
                        "#922b3e": "Red-violet (Color wheel)",
                        "#a45a52": "Redwood",
                        "#002387": "Resolution blue",
                        "#777696": "Rhythm",
                        "#004040": "Rich black",
                        "#010b13": "Rich black (FOGRA29)",
                        "#010203": "Rich black (FOGRA39)",
                        "#444c38": "Rifle green",
                        "#00cccc": "Robin egg blue",
                        "#8a7f80": "Rocket metallic",
                        "#838996": "Roman silver",
                        "#ff007f": "Rose",
                        "#f9429e": "Rose bonbon",
                        "#9e5e6f": "Rose Dust",
                        "#674846": "Rose ebony",
                        "#e32636": "Rose madder",
                        "#ff66cc": "Rose pink",
                        "#c21e56": "Rose red",
                        "#905d5d": "Rose taupe",
                        "#ab4e52": "Rose vale",
                        "#65000b": "Rosewood",
                        "#d40000": "Rosso corsa",
                        "#bc8f8f": "Rosy brown",
                        "#002366": "Royal blue (dark)",
                        "#4169e1": "Royal blue (light)",
                        "#7851a9": "Royal purple",
                        "#fada5e": "Royal yellow",
                        "#ce4676": "Ruber",
                        "#d10056": "Rubine red",
                        "#e0115f": "Ruby",
                        "#9b111e": "Ruby red",
                        "#a81c07": "Rufous",
                        "#80461b": "Russet",
                        "#679267": "Russian green",
                        "#32174d": "Russian violet",
                        "#b7410e": "Rust",
                        "#da2c43": "Rusty red",
                        "#043927": "Sacramento State green",
                        "#8b4513": "Saddle brown",
                        "#ff7800": "Safety orange",
                        "#ff6700": "Safety orange (blaze orange)",
                        "#eed202": "Safety yellow",
                        "#f4c430": "Saffron",
                        "#bcb88a": "Sage",
                        "#23297a": "St. Patrick's blue",
                        "#fa8072": "Salmon",
                        "#ff91a4": "Salmon pink",
                        "#f4a460": "Sandy brown",
                        "#507d2a": "Sap green",
                        "#0f52ba": "Sapphire",
                        "#0067a5": "Sapphire blue",
                        "#cba135": "Satin sheen gold",
                        "#ff2400": "Scarlet",
                        "#ffd800": "School bus yellow",
                        "#66ff66": "Screamin' Green",
                        "#2e8b57": "Sea green",
                        "#00ffcd": "Sea green (Crayola)",
                        "#59260b": "Seal brown",
                        "#fff5ee": "Seashell",
                        "#ffba00": "Selective yellow",
                        "#704214": "Sepia",
                        "#8a795d": "Shadow",
                        "#778ba5": "Shadow blue",
                        "#009e60": "Shamrock green",
                        "#8fd400": "Sheen green",
                        "#d98695": "Shimmering Blush",
                        "#5fa778": "Shiny Shamrock",
                        "#fc0fc0": "Shocking pink",
                        "#882d17": "Sienna",
                        "#c0c0c0": "Silver",
                        "#aaa9ad": "Silver (Metallic)",
                        "#acacac": "Silver chalice",
                        "#c4aead": "Silver pink",
                        "#bfc1c2": "Silver sand",
                        "#cb410b": "Sinopia",
                        "#ff3855": "Sizzling Red",
                        "#ffdb00": "Sizzling Sunrise",
                        "#007474": "Skobeloff",
                        "#87ceeb": "Sky blue",
                        "#76d7ea": "Sky blue (Crayola)",
                        "#cf71af": "Sky magenta",
                        "#6a5acd": "Slate blue",
                        "#708090": "Slate gray",
                        "#299617": "Slimy green",
                        "#c84186": "Smitten",
                        "#100c08": "Smoky black",
                        "#fffafa": "Snow",
                        "#893843": "Solid pink",
                        "#757575": "Sonic silver",
                        "#1d2951": "Space cadet",
                        "#807532": "Spanish bistre",
                        "#0070b8": "Spanish blue",
                        "#d10047": "Spanish carmine",
                        "#989898": "Spanish gray",
                        "#009150": "Spanish green",
                        "#e86100": "Spanish orange",
                        "#f7bfbe": "Spanish pink",
                        "#e60026": "Spanish red",
                        "#4c2882": "Spanish violet",
                        "#007f5c": "Spanish viridian",
                        "#a7fc00": "Spring bud",
                        "#87ff2a": "Spring Frost",
                        "#00ff7f": "Spring green",
                        "#ecebbd": "Spring green (Crayola)",
                        "#007bb8": "Star command blue",
                        "#4682b4": "Steel blue",
                        "#cc33cc": "Steel pink",
                        "#5f8a8b": "Steel Teal",
                        "#e4d96f": "Straw",
                        "#914e75": "Sugar Plum",
                        "#ffcc33": "Sunglow",
                        "#e3ab57": "Sunray",
                        "#fad6a5": "Sunset",
                        "#cf6ba9": "Super pink",
                        "#a83731": "Sweet Brown",
                        "#d2b48c": "Tan",
                        "#d99a6c": "Tan (Crayola)",
                        "#f28500": "Tangerine",
                        "#fb4d46": "Tart Orange",
                        "#483c32": "Taupe",
                        "#8b8589": "Taupe gray",
                        "#d0f0c0": "Tea green",
                        "#f88379": "Tea rose",
                        "#008080": "Teal",
                        "#367588": "Teal blue",
                        "#cf3476": "Telemagenta",
                        "#cd5700": "Tenné (tawny)",
                        "#e2725b": "Terra cotta",
                        "#d8bfd8": "Thistle",
                        "#fc89ac": "Tickle Me Pink",
                        "#0abab5": "Tiffany Blue",
                        "#dbd7d2": "Timberwolf",
                        "#eee600": "Titanium yellow",
                        "#ff6347": "Tomato",
                        "#00755e": "Tropical rain forest",
                        "#2d68c4": "True Blue",
                        "#1c05b3": "Trypan Blue",
                        "#3e8ede": "Tufts blue",
                        "#deaa88": "Tumbleweed",
                        "#40e0d0": "Turquoise",
                        "#00ffef": "Turquoise blue",
                        "#a0d6b4": "Turquoise green",
                        "#7c4848": "Tuscan red",
                        "#c09999": "Tuscany",
                        "#8a496b": "Twilight lavender",
                        "#66023c": "Tyrian purple",
                        "#0033aa": "UA blue",
                        "#d9004c": "UA red",
                        "#3f00ff": "Ultramarine",
                        "#4166f5": "Ultramarine blue",
                        "#ff6fff": "Ultra pink",
                        "#fc6c85": "Ultra red",
                        "#635147": "Umber",
                        "#ffddca": "Unbleached silk",
                        "#5b92e5": "United Nations blue",
                        "#7b1113": "UP maroon",
                        "#ae2029": "Upsdell red",
                        "#afdbf5": "Uranian blue",
                        "#004f98": "USAFA blue",
                        "#664228": "Van Dyke brown",
                        "#f3e5ab": "Vanilla",
                        "#f38fa9": "Vanilla ice",
                        "#c5b358": "Vegas gold",
                        "#c80815": "Venetian red",
                        "#43b3ae": "Verdigris",
                        "#d9381e": "Vermilion1",
                        "#8f00ff": "Violet",
                        "#7f00ff": "Violet (color wheel)",
                        "#963d7f": "Violet (crayola)",
                        "#8601af": "Violet (RYB)",
                        "#ee82ee": "Violet (web)",
                        "#324ab2": "Violet-blue",
                        "#766ec8": "Violet-blue (Crayola)",
                        "#f75394": "Violet-red",
                        "#40826d": "Viridian",
                        "#009698": "Viridian green",
                        "#9f1d35": "Vivid burgundy",
                        "#00ccff": "Vivid sky blue",
                        "#ffa089": "Vivid tangerine",
                        "#9f00ff": "Vivid violet",
                        "#ceff00": "Volt",
                        "#004242": "Warm black",
                        "#f5deb3": "Wheat",
                        "#ffffff": "White",
                        "#a2add0": "Wild blue yonder",
                        "#d470a2": "Wild orchid",
                        "#ff43a4": "Wild Strawberry",
                        "#a75502": "Windsor tan",
                        "#722f37": "Wine",
                        "#673147": "Wine dregs",
                        "#ff007c": "Winter Sky",
                        "#56887d": "Wintergreen Dream",
                        "#c9a0dc": "Wisteria",
                        "#c19a6b": "Wood brown",
                        "#eeed09": "Xanthic",
                        "#738678": "Xanadu",
                        "#0c020f": "Xiketic",
                        "#0f4d92": "Yale Blue",
                        "#ffff00": "Yellow",
                        "#fce883": "Yellow (Crayola)",
                        "#efcc00": "Yellow (Munsell)",
                        "#ffd300": "Yellow (NCS)",
                        "#fedf00": "Yellow (Pantone)",
                        "#fefe33": "Yellow (RYB)",
                        "#9acd32": "Yellow-green",
                        "#c5e384": "Yellow-green (Crayola)",
                        "#30b21a": "Yellow-green (Color Wheel)",
                        "#ffae42": "Yellow Orange",
                        "#ff9505": "Yellow Orange (Color Wheel)",
                        "#2e5090": "YInMn Blue",
                        "#0014a8": "Zaffre",
                        "#39a78e": "Zomp"}
    
    def closest_colour(self, requested_colour):
        """
        Credits go to fraxel: https://stackoverflow.com/a/9694246/4040598
        :param requested_colour: #rrggbb string
        :return: The closest colour name as a string
        """
        requested_colour = hex_to_rgb(requested_colour)
        min_colours = {}
        for key, name in self.colours.items():
            r_c, g_c, b_c = hex_to_rgb(key)
            rd = (r_c - requested_colour[0]) ** 2
            gd = (g_c - requested_colour[1]) ** 2
            bd = (b_c - requested_colour[2]) ** 2
            min_colours[(rd + gd + bd)] = name

        c_name = min_colours[min(min_colours.keys())]
        hex_value = None
        for key, name in self.colours.items():
            if name == c_name:
                hex_value = ' {}'.format(key)
                break

        return c_name + hex_value

    def get_colour_name(self, requested_colour):
        try:
            closest_name = actual_name = self.colours[requested_colour]
        except KeyError:
            closest_name = self.closest_colour(requested_colour)
            actual_name = None

        return actual_name, closest_name




# -*- coding: utf-8 -*-
"""
    colorthief
    ~~~~~~~~~~

    Grabbing the color palette from an image.

    :copyright: (c) 2015 by Shipeng Feng.
    :license: BSD, see LICENSE for more details.
"""
__version__ = '0.2.1'

import math

from PIL import Image


class cached_property(object):
    """Decorator that creates converts a method with a single
    self argument into a property cached on the instance.
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, type):
        res = instance.__dict__[self.func.__name__] = self.func(instance)
        return res


class ColorThief(object):
    """Color thief main class."""

    def __init__(self, file):
        """Create one color thief for one image.

        :param file: A filename (string) or a file object. The file object
                     must implement `read()`, `seek()`, and `tell()` methods,
                     and be opened in binary mode.
        """
        self.image = Image.open(file)

    def get_color(self, quality=10):
        """Get the dominant color.

        :param quality: quality settings, 1 is the highest quality, the bigger
                        the number, the faster a color will be returned but
                        the greater the likelihood that it will not be the
                        visually most dominant color
        :return tuple: (r, g, b)
        """
        palette = self.get_palette(5, quality)
        return palette[0]

    def get_palette(self, color_count=10, quality=10):
        """Build a color palette.  We are using the median cut algorithm to
        cluster similar colors.

        :param color_count: the size of the palette, max number of colors
        :param quality: quality settings, 1 is the highest quality, the bigger
                        the number, the faster the palette generation, but the
                        greater the likelihood that colors will be missed.
        :return list: a list of tuple in the form (r, g, b)
        """
        image = self.image.convert('RGBA')
        width, height = image.size
        pixels = image.getdata()
        pixel_count = width * height
        valid_pixels = []
        for i in range(0, pixel_count, quality):
            r, g, b, a = pixels[i]
            # If pixel is mostly opaque and not white
            if a >= 125:
                if not (r > 250 and g > 250 and b > 250):
                    valid_pixels.append((r, g, b))

        # Send array to quantize function which clusters values
        # using median cut algorithm
        cmap = MMCQ.quantize(valid_pixels, color_count)
        return cmap.palette


class MMCQ(object):
    """Basic Python port of the MMCQ (modified median cut quantization)
    algorithm from the Leptonica library (http://www.leptonica.com/).
    """

    SIGBITS = 5
    RSHIFT = 8 - SIGBITS
    MAX_ITERATION = 1000
    FRACT_BY_POPULATIONS = 0.75

    @staticmethod
    def get_color_index(r, g, b):
        return (r << (2 * MMCQ.SIGBITS)) + (g << MMCQ.SIGBITS) + b

    @staticmethod
    def get_histo(pixels):
        """histo (1-d array, giving the number of pixels in each quantized
        region of color space)
        """
        histo = dict()
        for pixel in pixels:
            rval = pixel[0] >> MMCQ.RSHIFT
            gval = pixel[1] >> MMCQ.RSHIFT
            bval = pixel[2] >> MMCQ.RSHIFT
            index = MMCQ.get_color_index(rval, gval, bval)
            histo[index] = histo.setdefault(index, 0) + 1
        return histo

    @staticmethod
    def vbox_from_pixels(pixels, histo):
        rmin = 1000000
        rmax = 0
        gmin = 1000000
        gmax = 0
        bmin = 1000000
        bmax = 0
        for pixel in pixels:
            rval = pixel[0] >> MMCQ.RSHIFT
            gval = pixel[1] >> MMCQ.RSHIFT
            bval = pixel[2] >> MMCQ.RSHIFT
            rmin = min(rval, rmin)
            rmax = max(rval, rmax)
            gmin = min(gval, gmin)
            gmax = max(gval, gmax)
            bmin = min(bval, bmin)
            bmax = max(bval, bmax)
        return VBox(rmin, rmax, gmin, gmax, bmin, bmax, histo)

    @staticmethod
    def median_cut_apply(histo, vbox):
        if not vbox.count:
            return (None, None)

        rw = vbox.r2 - vbox.r1 + 1
        gw = vbox.g2 - vbox.g1 + 1
        bw = vbox.b2 - vbox.b1 + 1
        maxw = max([rw, gw, bw])
        # only one pixel, no split
        if vbox.count == 1:
            return (vbox.copy, None)
        # Find the partial sum arrays along the selected axis.
        total = 0
        sum_ = 0
        partialsum = {}
        lookaheadsum = {}
        do_cut_color = None
        if maxw == rw:
            do_cut_color = 'r'
            for i in range(vbox.r1, vbox.r2 + 1):
                sum_ = 0
                for j in range(vbox.g1, vbox.g2 + 1):
                    for k in range(vbox.b1, vbox.b2 + 1):
                        index = MMCQ.get_color_index(i, j, k)
                        sum_ += histo.get(index, 0)
                total += sum_
                partialsum[i] = total
        elif maxw == gw:
            do_cut_color = 'g'
            for i in range(vbox.g1, vbox.g2 + 1):
                sum_ = 0
                for j in range(vbox.r1, vbox.r2 + 1):
                    for k in range(vbox.b1, vbox.b2 + 1):
                        index = MMCQ.get_color_index(j, i, k)
                        sum_ += histo.get(index, 0)
                total += sum_
                partialsum[i] = total
        else:  # maxw == bw
            do_cut_color = 'b'
            for i in range(vbox.b1, vbox.b2 + 1):
                sum_ = 0
                for j in range(vbox.r1, vbox.r2 + 1):
                    for k in range(vbox.g1, vbox.g2 + 1):
                        index = MMCQ.get_color_index(j, k, i)
                        sum_ += histo.get(index, 0)
                total += sum_
                partialsum[i] = total
        for i, d in partialsum.items():
            lookaheadsum[i] = total - d

        # determine the cut planes
        dim1 = do_cut_color + '1'
        dim2 = do_cut_color + '2'
        dim1_val = getattr(vbox, dim1)
        dim2_val = getattr(vbox, dim2)
        for i in range(dim1_val, dim2_val + 1):
            if partialsum[i] > (total / 2):
                vbox1 = vbox.copy
                vbox2 = vbox.copy
                left = i - dim1_val
                right = dim2_val - i
                if left <= right:
                    d2 = min([dim2_val - 1, int(i + right / 2)])
                else:
                    d2 = max([dim1_val, int(i - 1 - left / 2)])
                # avoid 0-count boxes
                while not partialsum.get(d2, False):
                    d2 += 1
                count2 = lookaheadsum.get(d2)
                while not count2 and partialsum.get(d2 - 1, False):
                    d2 -= 1
                    count2 = lookaheadsum.get(d2)
                # set dimensions
                setattr(vbox1, dim2, d2)
                setattr(vbox2, dim1, getattr(vbox1, dim2) + 1)
                return (vbox1, vbox2)
        return (None, None)

    @staticmethod
    def quantize(pixels, max_color):
        """Quantize.

        :param pixels: a list of pixel in the form (r, g, b)
        :param max_color: max number of colors
        """
        if not pixels:
            raise Exception('Empty pixels when quantize.')
        if max_color < 2 or max_color > 256:
            raise Exception('Wrong number of max colors when quantize.')

        histo = MMCQ.get_histo(pixels)

        # check that we aren't below maxcolors already
        if len(histo) <= max_color:
            # generate the new colors from the histo and return
            pass

        # get the beginning vbox from the colors
        vbox = MMCQ.vbox_from_pixels(pixels, histo)
        pq = PQueue(lambda x: x.count)
        pq.push(vbox)

        # inner function to do the iteration
        def iter_(lh, target):
            n_color = 1
            n_iter = 0
            while n_iter < MMCQ.MAX_ITERATION:
                vbox = lh.pop()
                if not vbox.count:  # just put it back
                    lh.push(vbox)
                    n_iter += 1
                    continue
                # do the cut
                vbox1, vbox2 = MMCQ.median_cut_apply(histo, vbox)
                if not vbox1:
                    raise Exception("vbox1 not defined; shouldn't happen!")
                lh.push(vbox1)
                if vbox2:  # vbox2 can be null
                    lh.push(vbox2)
                    n_color += 1
                if n_color >= target:
                    return
                if n_iter > MMCQ.MAX_ITERATION:
                    return
                n_iter += 1

        # first set of colors, sorted by population
        iter_(pq, MMCQ.FRACT_BY_POPULATIONS * max_color)

        # Re-sort by the product of pixel occupancy times the size in
        # color space.
        pq2 = PQueue(lambda x: x.count * x.volume)
        while pq.size():
            pq2.push(pq.pop())

        # next set - generate the median cuts using the (npix * vol) sorting.
        iter_(pq2, max_color - pq2.size())

        # calculate the actual colors
        cmap = CMap()
        while pq2.size():
            cmap.push(pq2.pop())
        return cmap


class VBox(object):
    """3d color space box"""

    def __init__(self, r1, r2, g1, g2, b1, b2, histo):
        self.r1 = r1
        self.r2 = r2
        self.g1 = g1
        self.g2 = g2
        self.b1 = b1
        self.b2 = b2
        self.histo = histo

    @cached_property
    def volume(self):
        sub_r = self.r2 - self.r1
        sub_g = self.g2 - self.g1
        sub_b = self.b2 - self.b1
        return (sub_r + 1) * (sub_g + 1) * (sub_b + 1)

    @property
    def copy(self):
        return VBox(self.r1, self.r2, self.g1, self.g2,
                    self.b1, self.b2, self.histo)

    @cached_property
    def avg(self):
        ntot = 0
        mult = 1 << (8 - MMCQ.SIGBITS)
        r_sum = 0
        g_sum = 0
        b_sum = 0
        for i in range(self.r1, self.r2 + 1):
            for j in range(self.g1, self.g2 + 1):
                for k in range(self.b1, self.b2 + 1):
                    histoindex = MMCQ.get_color_index(i, j, k)
                    hval = self.histo.get(histoindex, 0)
                    ntot += hval
                    r_sum += hval * (i + 0.5) * mult
                    g_sum += hval * (j + 0.5) * mult
                    b_sum += hval * (k + 0.5) * mult

        if ntot:
            r_avg = int(r_sum / ntot)
            g_avg = int(g_sum / ntot)
            b_avg = int(b_sum / ntot)
        else:
            r_avg = int(mult * (self.r1 + self.r2 + 1) / 2)
            g_avg = int(mult * (self.g1 + self.g2 + 1) / 2)
            b_avg = int(mult * (self.b1 + self.b2 + 1) / 2)

        return r_avg, g_avg, b_avg

    def contains(self, pixel):
        rval = pixel[0] >> MMCQ.RSHIFT
        gval = pixel[1] >> MMCQ.RSHIFT
        bval = pixel[2] >> MMCQ.RSHIFT
        return all([
            rval >= self.r1,
            rval <= self.r2,
            gval >= self.g1,
            gval <= self.g2,
            bval >= self.b1,
            bval <= self.b2,
        ])

    @cached_property
    def count(self):
        npix = 0
        for i in range(self.r1, self.r2 + 1):
            for j in range(self.g1, self.g2 + 1):
                for k in range(self.b1, self.b2 + 1):
                    index = MMCQ.get_color_index(i, j, k)
                    npix += self.histo.get(index, 0)
        return npix


class CMap(object):
    """Color map"""

    def __init__(self):
        self.vboxes = PQueue(lambda x: x['vbox'].count * x['vbox'].volume)

    @property
    def palette(self):
        return self.vboxes.map(lambda x: x['color'])

    def push(self, vbox):
        self.vboxes.push({
            'vbox': vbox,
            'color': vbox.avg,
        })

    def size(self):
        return self.vboxes.size()

    def nearest(self, color):
        d1 = None
        p_color = None
        for i in range(self.vboxes.size()):
            vbox = self.vboxes.peek(i)
            d2 = math.sqrt(
                math.pow(color[0] - vbox['color'][0], 2) +
                math.pow(color[1] - vbox['color'][1], 2) +
                math.pow(color[2] - vbox['color'][2], 2)
            )
            if d1 is None or d2 < d1:
                d1 = d2
                p_color = vbox['color']
        return p_color

    def map(self, color):
        for i in range(self.vboxes.size()):
            vbox = self.vboxes.peek(i)
            if vbox['vbox'].contains(color):
                return vbox['color']
        return self.nearest(color)


class PQueue(object):
    """Simple priority queue."""

    def __init__(self, sort_key):
        self.sort_key = sort_key
        self.contents = []
        self._sorted = False

    def sort(self):
        self.contents.sort(key=self.sort_key)
        self._sorted = True

    def push(self, o):
        self.contents.append(o)
        self._sorted = False

    def peek(self, index=None):
        if not self._sorted:
            self.sort()
        if index is None:
            index = len(self.contents) - 1
        return self.contents[index]

    def pop(self):
        if not self._sorted:
            self.sort()
        return self.contents.pop()

    def size(self):
        return len(self.contents)

    def map(self, f):
        return list(map(f, self.contents))



#!/usr/bin/env python3
# _*_ coding: utf-8 _*_

"""
Wallpaper and colour manager for Sway, i3 and some other WMs, as a frontend to swaybg and feh

Author: Piotr Miller & Contributors
e-mail: nwg.piotr@gmail.com
Project: https://github.com/nwg-piotr/azote
License: GPL3

"""
import os
import sys
import subprocess
import stat
import gi
import cairo
from PIL import Image
from azote import common

# send2trash module may or may not be available
try:
    from send2trash import send2trash

    common.env['send2trash'] = True
except Exception:
    common.env['send2trash'] = False
    print('python-send2trash package not found - deleting pictures unavailable')

from azote.colorthief import ColorThief

dir_name = os.path.dirname(__file__)

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GdkPixbuf, Gdk, GLib
from gi.repository.GdkPixbuf import InterpType
from azote.tools import set_env, hash_name, create_thumbnails, file_allowed, update_status_bar, flip_selected_wallpaper, \
    copy_backgrounds, create_pixbuf, split_selected_wallpaper, scale_and_crop, clear_thumbnails, current_display, \
    save_json, load_json
from azote.color_tools import rgba_to_hex, hex_to_rgb, rgb_to_hex, rgb_to_rgba
from azote.plugins import Alacritty, Xresources
from azote.color_tools import WikiColours

try:
    gi.require_version('AppIndicator3', '0.1')
    from gi.repository import AppIndicator3

    common.env['app_indicator'] = True
except:
    common.env['app_indicator'] = False
    print('libappindicator-gtk3 package not found - tray icon unavailable')

from azote.__about__ import __version__

def get_files():
    try:
        inames = "-iname \"*."+"\" -o -iname \"*.".join(common.allowed_file_types)+"\""
        files = subprocess.check_output("find '%s' -mindepth 1 %s" %(common.settings.src_path, inames), shell=True).decode().split("\n")[:-1]
        file_names = [f.replace(common.settings.src_path+"/", "") for f in files]
    except FileNotFoundError:
        common.settings.src_path = os.getenv('HOME')
        file_names = [f for f in os.listdir(common.settings.src_path)
                      if os.path.isfile(os.path.join(common.settings.src_path, f))]

    if common.settings.sorting == 'new':
        file_names.sort(reverse=True, key=lambda f: os.path.getmtime(os.path.join(common.settings.src_path, f)))
    elif common.settings.sorting == 'old':
        file_names.sort(key=lambda f: os.path.getmtime(os.path.join(common.settings.src_path, f)))
    elif common.settings.sorting == 'az':
        file_names.sort()
    elif common.settings.sorting == 'za':
        file_names.sort(reverse=True)

    return file_names


class Preview(Gtk.ScrolledWindow):
    def __init__(self):
        super().__init__()

        self.set_border_width(10)
        self.set_propagate_natural_height(True)
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS)

        common.thumbnails_list = []
        self.grid = Gtk.FlowBox()
        self.grid.set_valign(Gtk.Align.START)
        # self.grid.set_max_children_per_line(30)
        self.grid.set_selection_mode(Gtk.SelectionMode.NONE)

        create_thumbnails(common.settings.src_path)
        self.files_dict = dict([(f, None) for f in os.listdir(common.settings.src_path)])

        src_pictures = get_files()

        for file in src_pictures:
            if file_allowed(file):
                thumbnail = Thumbnail(common.settings.src_path, file)
                common.thumbnails_list.append(thumbnail)
                self.grid.add(thumbnail)

        self.add(self.grid)

    def refresh(self, create_thumbs=True):
        self.files_dict = dict([(f, None) for f in os.listdir(common.settings.src_path)])

        if create_thumbs:
            create_thumbnails(common.settings.src_path)

        for thumbnail in common.thumbnails_list:
            self.grid.remove(thumbnail)
            thumbnail.destroy()

        src_pictures = get_files()

        for file in src_pictures:
            if file_allowed(file):
                thumbnail = Thumbnail(common.settings.src_path, file)
                common.thumbnails_list.append(thumbnail)
                self.grid.add(thumbnail)

                thumbnail.show_all()
                thumbnail.toolbar.hide()

        update_status_bar()


class Thumbnail(Gtk.VBox):
    def __init__(self, folder, filename):
        super().__init__()
        self.toolbar = ImageToolbar(self)
        self.add(self.toolbar)

        self.image_button = Gtk.Button()
        self.image_button.set_property("name", "thumb-btn")
        self.image_button.set_always_show_image(True)

        self.folder = folder
        self.filename = filename
        self.source_path = os.path.join(folder, filename)

        self.img = Gtk.Image()
        self.thumb_file = "{}.png".format(os.path.join(common.thumb_dir, hash_name(self.source_path)))
        self.img.set_from_file(self.thumb_file)

        self.image_button.set_image(self.img)
        self.image_button.set_image_position(2)  # TOP
        self.image_button.set_tooltip_text(common.lang['thumbnail_tooltip'])

        if len(filename) > 30:
            filename = '…{}'.format(filename[-28::])
        self.image_button.set_label(filename)
        self.selected = False

        # self.connect('clicked', self.on_button_press)
        self.image_button.connect('button-press-event', self.on_image_button_press)

        self.add(self.image_button)

    def on_image_button_press(self, button, event):

        self.select(button)

        if event.type == Gdk.EventType._2BUTTON_PRESS:
            on_thumb_double_click(button)
        if event.button == 3:
            show_image_menu(self)

    def on_menu_button_press(self, button):
        show_image_menu(self)

    def select(self, thumbnail):
        if common.split_button:
            common.split_button.set_sensitive(True)

        common.apply_to_all_button.set_sensitive(True)

        self.image_button.selected = True
        common.selected_wallpaper = self
        deselect_all()
        if common.settings.image_menu_button:
            self.toolbar.show_all()
        thumbnail.set_property("name", "thumb-btn-selected")

        with Image.open(self.source_path) as img:
            filename = self.filename
            if len(filename) > 30:
                filename = '…{}'.format(filename[-28::])
            common.selected_picture_label.set_text("{} ({} x {})".format(filename, img.size[0], img.size[1]))

    def deselect(self, thumbnail):
        self.selected = False
        self.toolbar.hide()
        thumbnail.image_button.set_property("name", "thumb-btn")


def deselect_all():
    for thumbnail in common.thumbnails_list:
        thumbnail.deselect(thumbnail)


class ImageToolbar(Gtk.HBox):
    def __init__(self, thumbnail):
        super().__init__()
        self.parent_thumbnail = thumbnail
        self.set_spacing(0)
        self.set_border_width(0)

        test_label = Gtk.Label()
        test_label.set_text('')
        self.pack_start(test_label, True, True, 0)

        self.menu_btn = Gtk.EventBox()
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_image_menu.svg'))
        self.menu_btn.add(img)
        self.menu_btn.connect('button-press-event', self.on_menu_button_press)
        self.pack_start(self.menu_btn, False, False, 0)

    def on_menu_button_press(self, button, event):
        show_image_menu(self.parent_thumbnail, from_toolbar=True)


class DisplayBox(Gtk.Box):
    """
    The box contains elements to preview certain displays and assign wallpapers to them
    """

    def __init__(self, name, width, height, path=None, thumb=None, xrandr_idx=None):
        super().__init__()

        self.set_orientation(Gtk.Orientation.VERTICAL)

        # Values to assigned to corresponding display when apply button pressed
        self.display_name = name
        self.wallpaper_path = path
        self.thumbnail_path = thumb
        self.mode = 'fill' if common.sway or common.env['wayland'] else 'scale'
        self.color = None
        self.xrandr_idx = xrandr_idx
        self.include = True

        if thumb and os.path.isfile(thumb):
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(thumb)
        else:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(os.path.join(dir_name, 'images/empty.png'))
        pixbuf = pixbuf.scale_simple(common.settings.thumb_size[0], common.settings.thumb_size[1], InterpType.BILINEAR)

        self.img = Gtk.Image.new_from_pixbuf(pixbuf)

        self.select_button = Gtk.Button()
        self.select_button.set_always_show_image(True)
        self.select_button.set_label("{} ({} x {})".format(name, width, height))  # label on top: name (with x height)
        self.select_button.set_image(self.img)  # preview of selected wallpaper
        self.select_button.set_image_position(3)  # label on top, image below
        self.select_button.set_property("name", "display-btn")  # to assign css style
        self.select_button.set_tooltip_text(common.lang['set_selected_wallpaper'])

        self.pack_start(self.select_button, False, False, 10)

        self.select_button.connect_after('clicked', self.on_select_button)

        # Combo box to choose a mode to use for the image
        mode_selector = Gtk.ListStore(str)

        if common.sway or common.env['wayland']:
            for mode in common.modes_swaybg:
                mode_selector.append([mode])
        else:
            for mode in common.modes_feh:
                mode_selector.append([mode])

        # Let's display the mode combo and the color button side-by-side in a vertical box
        options_box = Gtk.Box()
        options_box.set_spacing(15)
        options_box.set_border_width(0)
        options_box.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.pack_start(options_box, True, False, 0)

        check_button = Gtk.CheckButton()
        check_button.set_active(self.include)
        check_button.set_tooltip_text(common.lang["include_when_splitting"])
        check_button.connect("toggled", self.switch_included)
        options_box.pack_start(check_button, False, False, 0)

        self.mode_combo = Gtk.ComboBox.new_with_model(mode_selector)
        self.mode_combo.set_active(2)
        self.mode_combo.connect("changed", self.on_mode_combo_changed)
        renderer_text = Gtk.CellRendererText()
        self.mode_combo.pack_start(renderer_text, True)
        self.mode_combo.add_attribute(renderer_text, "text", 0)
        self.mode_combo.set_tooltip_text(common.lang['display_mode'])
        options_box.add(self.mode_combo)

        if common.sway or common.env['wayland']:
            # Color button
            self.color_button = Gtk.ColorButton()
            color = Gdk.RGBA()
            color.red = 0.0
            color.green = 0.0
            color.blue = 0.0
            color.alpha = 1.0
            self.color_button.set_rgba(color)
            self.color_button.connect("color-set", self.on_color_chosen, self.color_button)
            self.color_button.set_tooltip_text(common.lang['background_color'])
            options_box.add(self.color_button)

        self.flip_button = Gtk.Button()
        self.flip_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_flip.svg'))
        self.flip_button.set_image(img)
        self.flip_button.set_tooltip_text(common.lang['flip_image'])
        self.flip_button.set_sensitive(False)
        self.flip_button.connect('clicked', self.on_flip_button)
        self.flip_button.set_tooltip_text(common.lang['flip_wallpaper_horizontally'])
        options_box.pack_start(self.flip_button, True, True, 0)

    def switch_included(self, ckb):
        self.include = ckb.get_active()

    def clear_color_selection(self):
        # If not on sway / swaybg, we have no color_button in UI
        if common.sway or common.env['wayland']:
            # clear color selection: image will be used
            color = Gdk.RGBA()
            color.red = 0.0
            color.green = 0.0
            color.blue = 0.0
            color.alpha = 1.0
            self.color_button.set_rgba(color)
            self.color = None

    def on_select_button(self, button):
        if common.selected_wallpaper:
            self.img.set_from_file(common.selected_wallpaper.thumb_file)
            self.wallpaper_path = common.selected_wallpaper.source_path
            self.thumbnail_path = common.selected_wallpaper.thumb_file
            button.set_property("name", "display-btn-selected")
            self.flip_button.set_sensitive(True)

            self.clear_color_selection()

            common.apply_button.set_sensitive(True)

    def on_mode_combo_changed(self, combo):
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            model = combo.get_model()
            mode = model[tree_iter][0]
            self.mode = mode

        # If our backend is feh, not swaybg, we can not set mode for each wallpaper separately.
        # Let's copy the same selection to all displays.
        if not common.sway and common.env['wayland'] and common.display_boxes_list:
            selection = combo.get_active()
            for box in common.display_boxes_list:
                box.mode_combo.set_active(selection)

    def on_color_chosen(self, user_data, button):
        self.color = rgba_to_hex(button.get_rgba())
        # clear selected image to indicate it won't be used
        self.img.set_from_file(os.path.join(dir_name, "images/empty.png"))
        common.apply_button.set_sensitive(True)

    def on_flip_button(self, button):
        # convert images and get (thumbnail path, flipped image path)
        images = flip_selected_wallpaper()
        self.img.set_from_file(images[0])
        self.wallpaper_path = images[1]
        self.thumbnail_path = images[0]
        self.flip_button.set_sensitive(False)


class SortingButton(Gtk.Button):
    def __init__(self):
        super().__init__()
        self.set_always_show_image(True)
        self.img = Gtk.Image()
        self.refresh()
        self.set_tooltip_text(common.lang['sorting_order'])
        self.connect('clicked', self.on_sorting_button)

    def refresh(self):
        if common.settings.sorting == 'old':
            self.img.set_from_file(os.path.join(dir_name, 'images/icon_old.svg'))
        elif common.settings.sorting == 'az':
            self.img.set_from_file(os.path.join(dir_name, 'images/icon_az.svg'))
        elif common.settings.sorting == 'za':
            self.img.set_from_file(os.path.join(dir_name, 'images/icon_za.svg'))
        else:
            self.img.set_from_file(os.path.join(dir_name, 'images/icon_new.svg'))
        self.set_image(self.img)

    def on_sorting_button(self, widget):
        menu = Gtk.Menu()
        i0 = Gtk.MenuItem.new_with_label(common.lang['sorting_new'])
        i0.connect('activate', self.on_i0)
        menu.append(i0)
        i1 = Gtk.MenuItem.new_with_label(common.lang['sorting_old'])
        i1.connect('activate', self.on_i1)
        menu.append(i1)
        i2 = Gtk.MenuItem.new_with_label(common.lang['sorting_az'])
        i2.connect('activate', self.on_i2)
        menu.append(i2)
        i3 = Gtk.MenuItem.new_with_label(common.lang['sorting_za'])
        i3.connect('activate', self.on_i3)
        menu.append(i3)
        menu.show_all()
        menu.popup_at_widget(widget, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH_WEST, None)

    def on_i0(self, widget):
        common.settings.sorting = 'new'
        common.settings.save()
        self.refresh()
        common.preview.refresh()

    def on_i1(self, widget):
        common.settings.sorting = 'old'
        common.settings.save()
        self.refresh()
        common.preview.refresh()

    def on_i2(self, widget):
        common.settings.sorting = 'az'
        common.settings.save()
        self.refresh()
        common.preview.refresh()

    def on_i3(self, widget):
        common.settings.sorting = 'za'
        common.settings.save()
        self.refresh()
        common.preview.refresh()


def on_apply_button(button):
    """
    Create the command for swaybg (Sway) or feh (X11)
    """
    # Copy modified wallpapers (if any) from temporary to backgrounds folder
    copy_backgrounds()

    if common.sway or common.env['wayland']:
        # On next startup we'll want to restore what we're setting here. Let's save it to json files
        # instead of parsing .azoteb / .fehbg. We only save pictures.
        restore_from_file = os.path.join(common.data_home, "swaybg.json")
        restore_from = []

        # Prepare, save and execute the shell script for swaybg. It'll be placed in ~/.azotebg for further use.
        batch_content = ['#!/usr/bin/env bash', 'pkill swaybg']
        for box in common.display_boxes_list:
            if box.color:
                # if a color chosen, the wallpaper won't appear
                batch_content.append("swaybg -o {} -c{} &".format(box.display_name, box.color))
            elif box.wallpaper_path:
                # see: https://github.com/nwg-piotr/azote/issues/143
                if common.settings.generic_display_names:
                    display_name = ""
                    for item in common.displays:
                        if item["name"] == box.display_name:
                            display_name = item["generic-name"]
                else:
                    display_name = box.display_name

                # Escape some special characters which would mess up the script
                wallpaper_path = box.wallpaper_path.replace('\\', '\\\\').replace("$", "\\$").replace("`",
                                                                                                     "\\`").replace('"',
                                                                                                                    '\\"')

                batch_content.append(
                    "swaybg -o '{}' -i \"{}\" -m {} &".format(display_name, wallpaper_path, box.mode))

                # build the json file content
                if box.wallpaper_path.startswith("{}/backgrounds-sway/flipped-".format(common.data_home)):
                    thumb = "thumbnail-{}".format(box.wallpaper_path.split("flipped-")[1])
                    thumb = os.path.join(common.data_home, "backgrounds-sway", thumb)

                elif box.wallpaper_path.startswith("{}/backgrounds-sway/part".format(common.data_home)):
                    thumb = "thumb-part{}".format(box.wallpaper_path.split("part")[1])
                    thumb = os.path.join(common.data_home, "backgrounds-sway", thumb)

                else:
                    thumb = "{}.png".format(hash_name(box.wallpaper_path))
                    thumb = os.path.join(common.data_home, "thumbnails", thumb)

                entry = {"name": box.display_name, "path": box.wallpaper_path, "thumb": thumb}
                restore_from.append(entry)

        with open(common.cmd_file, 'w') as f:
            for item in batch_content:
                f.write("%s\n" % item)
        # make the file executable
        st = os.stat(common.cmd_file)
        os.chmod(common.cmd_file, st.st_mode | stat.S_IEXEC)

        subprocess.call(common.cmd_file, shell=True)

        save_json(restore_from, restore_from_file)

    else:
        # As above
        restore_from_file = os.path.join(common.data_home, "feh.json")
        restore_from = []

        # Prepare and execute the feh command. It's being saved automagically to ~/.fehbg
        mode = common.display_boxes_list[0].mode  # They are all the same, just check the 1st one
        command = "feh --bg-{}".format(mode)
        c = common.display_boxes_list.copy()
        c = sorted(c, key=lambda x: x.xrandr_idx)
        for box in c:
            command += " '{}'".format(box.wallpaper_path)

            # build the json file content
            if box.wallpaper_path.startswith("{}/backgrounds-feh/flipped-".format(common.data_home)):
                thumb = "thumbnail-{}".format(box.wallpaper_path.split("flipped-")[1])
                thumb = os.path.join(common.data_home, "backgrounds-feh", thumb)

            elif box.wallpaper_path.startswith("{}/backgrounds-feh/part".format(common.data_home)):
                thumb = "thumb-part{}".format(box.wallpaper_path.split("part")[1])
                thumb = os.path.join(common.data_home, "backgrounds-feh", thumb)

            else:
                thumb = "{}.png".format(hash_name(box.wallpaper_path))
                thumb = os.path.join(common.data_home, "thumbnails", thumb)

            entry = {"name": box.display_name, "path": box.wallpaper_path,
                     "thumb": thumb}
            restore_from.append(entry)

        subprocess.call(command, shell=True)

        save_json(restore_from, restore_from_file)


def on_split_button(button):
    if common.selected_wallpaper:
        common.apply_button.set_sensitive(True)
        num_parts = 0
        for item in common.display_boxes_list:
            if item.include:
                num_parts += 1
        paths = split_selected_wallpaper(num_parts)
        i = 0
        for box in common.display_boxes_list:
            if box.include:
                box.wallpaper_path = paths[i][0]
                box.img.set_from_file(paths[i][1])
                box.thumbnail_path = paths[i][1]
                i += 1

    if common.display_boxes_list:
        for box in common.display_boxes_list:
            box.clear_color_selection()


def open_with(item, opener):
    # if feh selected as the opener, let's start it with options as below
    if opener == 'feh':
        command = 'feh --start-at "{}" --scale-down --no-fehbg -d --output-dir {}'.format(
            common.selected_wallpaper.source_path, common.selected_wallpaper.folder)
    # elif could specify options for other certain programs here
    elif opener == 'swappy':
        command = 'swappy -f {}'.format(common.selected_wallpaper.source_path)
    else:
        command = '{} "{}"'.format(opener, common.selected_wallpaper.source_path)
    subprocess.Popen(command, shell=True)


def clear_wallpaper_selection():
    common.selected_wallpaper = None
    common.selected_picture_label.set_text(common.lang['no_picture_selected'])
    if common.split_button:
        common.split_button.set_sensitive(False)
    common.apply_button.set_sensitive(False)
    common.apply_to_all_button.set_sensitive(False)


def on_about_button(button):
    dialog = Gtk.AboutDialog()
    dialog.set_program_name('Azote')

    try:
        # version = pkg_resources.require(common.app_name)[0].version
        dialog.set_version("v{}".format(__version__))
    except Exception as e:
        print("Couldn't check version: {}".format(e))
        pass

    logo = GdkPixbuf.Pixbuf.new_from_file_at_size(os.path.join(dir_name, 'images/azote.svg'), 96, 96)

    dialog.set_keep_above(True)
    dialog.set_logo(logo)
    dialog.set_copyright('(c) 2019-2023 Piotr Miller & Contributors')
    dialog.set_website('https://github.com/nwg-piotr/azote')
    dialog.set_comments(common.lang['app_desc'])
    dialog.set_license_type(Gtk.License.GPL_3_0)
    dialog.set_authors(['Piotr Miller (nwg) and Contributors', 'Libraries and dependencies:',
                        '- colorthief python module (c) 2015 Shipeng Feng',
                        '- python-pillow (c) 1995-2011, Fredrik Lundh, 2010-2019 Alex Clark and Contributors',
                        '- pygobject (c) 2005-2019 The GNOME Project',
                        '- GTK+ (c) 2007-2019 The GTK Team',
                        '- feh (c) 1999,2000 Tom Gilbert, 2010-2018 Daniel Friesel',
                        '- swaybg (c) 2016-2019 Drew DeVault',
                        '- wlr-randr (c) 2019 Purism SPC and Contributors',
                        '- send2trash python module (c) 2017 Virgil Dupras',
                        '- grim, slurp (c) 2018 emersion',
                        '- maim, slop (c) 2014 Dalton Nell and Contributors',
                        '- imagemagick (c) 1999-2019 ImageMagick Studio LLC',
                        '- PyYAML (c) 2017-2019 Ingy döt Net Copyright (c) 2006-2016 Kirill Simonov'])
    dialog.set_translator_credits('xsme, Leon-Plickat (de_DE), HumanG33k (fr_FR), Andrea Frati (it_IT),\nMarcelo dos Santos (pt_BR), Aleksey Samoilov (ru_RU)')
    dialog.set_artists(['edskeye'])

    dialog.show()

    dialog.run()
    dialog.destroy()
    return False


def move_to_trash(widget):
    send2trash(common.selected_wallpaper.source_path)
    if os.path.isfile(common.selected_wallpaper.thumb_file):
        send2trash(common.selected_wallpaper.thumb_file)
    clear_wallpaper_selection()
    common.preview.refresh()


def show_image_menu(widget, event=None, parent=None, from_toolbar=False):
    cd = current_display()
    if common.selected_wallpaper:
        if common.associations:  # not None if /usr/share/applications/mimeinfo.cache found and parse
            openers = common.associations[common.selected_wallpaper.source_path.split('.')[-1].lower()]
            menu = Gtk.Menu()
            if openers:
                for opener in openers:
                    # opener = (Name, Exec)
                    item = Gtk.MenuItem.new_with_label(common.lang['open_with'].format(opener[0]))
                    item.connect('activate', open_with, opener[1])
                    menu.append(item)
                item = Gtk.SeparatorMenuItem()
                menu.append(item)

            item = Gtk.MenuItem.new_with_label(common.lang['create_palette'])
            menu.append(item)
            submenu = Gtk.Menu()

            # Hell knows why the library does not return the tuple of expected length for some num_colors values
            # Let's cheat a little bit
            subitem = Gtk.MenuItem.new_with_label('6 {}'.format(common.lang['colors']))
            subitem.connect('activate', generate_palette, common.selected_wallpaper.thumb_file,
                            common.selected_wallpaper.filename,
                            common.selected_wallpaper.source_path, 6)
            submenu.append(subitem)

            subitem = Gtk.MenuItem.new_with_label('12 {}'.format(common.lang['colors']))
            subitem.connect('activate', generate_palette, common.selected_wallpaper.thumb_file,
                            common.selected_wallpaper.filename,
                            common.selected_wallpaper.source_path, 13)
            submenu.append(subitem)

            subitem = Gtk.MenuItem.new_with_label('18 {}'.format(common.lang['colors']))
            subitem.connect('activate', generate_palette, common.selected_wallpaper.thumb_file,
                            common.selected_wallpaper.filename,
                            common.selected_wallpaper.source_path, 19)
            submenu.append(subitem)

            subitem = Gtk.MenuItem.new_with_label('24 {}'.format(common.lang['colors']))
            subitem.connect('activate', generate_palette, common.selected_wallpaper.thumb_file,
                            common.selected_wallpaper.filename,
                            common.selected_wallpaper.source_path, 25)
            submenu.append(subitem)

            item.set_submenu(submenu)

            item = Gtk.MenuItem.new_with_label(common.lang['scale_and_crop'])
            menu.append(item)
            submenu = Gtk.Menu()

            # Scale and crop to detected displays dimension
            for i in range(len(common.displays)):
                display = common.displays[i]
                width, height = display['width'], display['height']
                subitem = Gtk.MenuItem.new_with_label(
                    '{} x {} ({})'.format(width, height, display['name']))
                subitem.connect('activate', scale_and_crop, common.selected_wallpaper.source_path, width, height)
                submenu.append(subitem)

            # Scale and crop to double width of the primary display
            display = common.displays[cd]
            width, height = display['width'] * 2, display['height']
            subitem = Gtk.MenuItem.new_with_label(
                '{} x {} ({} {} {})'.format(width, height, common.lang['current'], display['name'],
                                            common.lang['dual_width']))
            subitem.connect('activate', scale_and_crop, common.selected_wallpaper.source_path, width, height)
            submenu.append(subitem)

            # Scale and crop to double height of the primary display
            display = common.displays[cd]
            width, height = display['width'], display['height'] * 2
            subitem = Gtk.MenuItem.new_with_label(
                '{} x {} ({} {} {})'.format(width, height, common.lang['current'], display['name'],
                                            common.lang['dual_height']))
            subitem.connect('activate', scale_and_crop, common.selected_wallpaper.source_path, width, height)
            submenu.append(subitem)

            # Scale and crop to triple width of the primary display
            display = common.displays[cd]
            width, height = display['width'] * 3, display['height']
            subitem = Gtk.MenuItem.new_with_label(
                '{} x {} ({} {} {})'.format(width, height, common.lang['current'], display['name'],
                                            common.lang['triple_width']))
            subitem.connect('activate', scale_and_crop, common.selected_wallpaper.source_path, width, height)
            submenu.append(subitem)

            # Scale and crop to triple height of the primary display
            display = common.displays[cd]
            width, height = display['width'], display['height'] * 3
            subitem = Gtk.MenuItem.new_with_label(
                '{} x {} ({} {} {})'.format(width, height, common.lang['current'], display['name'],
                                            common.lang['triple_height']))
            subitem.connect('activate', scale_and_crop, common.selected_wallpaper.source_path, width, height)
            submenu.append(subitem)

            # Scale and crop to user-defined dimensions
            if common.settings.custom_display:
                subitem = Gtk.MenuItem.new_with_label(
                    '{} x {} ({})'.format(common.settings.custom_display[1], common.settings.custom_display[2],
                                          common.settings.custom_display[0]))
                subitem.connect('activate', scale_and_crop, common.selected_wallpaper.source_path,
                                int(common.settings.custom_display[1]), int(common.settings.custom_display[2]))
                submenu.append(subitem)

            item.set_submenu(submenu)

            if common.env['send2trash']:
                item = Gtk.SeparatorMenuItem()
                menu.append(item)
                item = Gtk.MenuItem.new_with_label(common.lang['remove_image'])
                menu.append(item)
                submenu = Gtk.Menu()
                item1 = Gtk.MenuItem.new_with_label(common.lang['move'])
                item1.connect('activate', move_to_trash)
                submenu.append(item1)
                item.set_submenu(submenu)

            menu.show_all()
            # We don't want the menu to stick out of the window on Sway, as it may be partially not clickable

            menu.popup_at_pointer(None)

        else:  # fallback in case mimeinfo.cache not found
            print("No registered program found. Does the /usr/share/applications/mimeinfo.cache file exist?")
            command = 'feh --start-at {} --scale-down --no-fehbg -d --output-dir {}'.format(
                common.selected_wallpaper.source_path, common.selected_wallpaper.folder)
            subprocess.Popen(command, shell=True)


def on_refresh_clicked(button):
    clear_wallpaper_selection()
    common.preview.refresh()


def generate_palette(item, thumb_file, filename, image_path, num_colors):
    color_thief = ColorThief(image_path)
    # dominant = color_thief.get_color(quality=10)
    palette = color_thief.get_palette(color_count=num_colors, quality=common.settings.palette_quality)
    if common.cpd:
        common.cpd.close()
    common.cpd = ColorPaletteDialog(thumb_file, filename, palette)


def on_folder_clicked(button):
    dialog = Gtk.FileChooserDialog(title=common.lang['open_folder'], parent=button.get_toplevel(),
                                   action=Gtk.FileChooserAction.SELECT_FOLDER)
    dialog.set_current_folder(common.settings.src_path)
    dialog.add_button(Gtk.STOCK_CANCEL, 0)
    dialog.add_button(Gtk.STOCK_OK, 1)
    dialog.set_default_response(1)
    dialog.set_default_size(800, 600)

    response = dialog.run()
    if response == 1:
        common.settings.src_path = dialog.get_filename()
        common.settings.save()
        dialog.destroy()
        common.preview.refresh()
        text = common.settings.src_path
        if len(text) > 40:
            text = '…{}'.format(text[-38::])
        button.set_label(text)

    dialog.destroy()
    clear_wallpaper_selection()


def destroy(self):
    Gtk.main_quit()


def check_height_and_start(window):
    w, h = window.get_size()
    window.destroy()
    if common.sway or common.env['wm'] == "i3":
        h = int(h * 0.95)
    print(
        "Available screen height: {} px; measurement delay: {} ms".format(h, common.settings.screen_measurement_delay))
    app = GUI(h)


class TransparentWindow(Gtk.Window):
    def __init__(self):
        Gtk.Window.__init__(self)
        self.connect('draw', self.draw)
        self.set_title("Checking...")
        # Credits for transparency go to KurtJacobson:
        # https://gist.github.com/KurtJacobson/374c8cb83aee4851d39981b9c7e2c22c
        screen = self.get_screen()
        visual = screen.get_rgba_visual()
        if visual and screen.is_composited():
            self.set_visual(visual)
        self.set_app_paintable(True)

    def draw(self, widget, context):
        context.set_source_rgba(0, 0, 0, 0.0)
        context.set_operator(cairo.OPERATOR_SOURCE)
        context.paint()
        context.set_operator(cairo.OPERATOR_OVER)


class GUI:
    def __init__(self, height):

        # set app_id for Wayland
        GLib.set_prgname('azote')

        window = Gtk.Window()
        h = height

        window.set_default_size(common.settings.thumb_width * (common.settings.columns + 1), h)
        common.main_window = window

        window.set_title("Azote~")
        window.set_icon_name("azote")
        window.set_role("azote")

        window.connect_after('destroy', destroy)
        window.connect("key-release-event", self.handle_keyboard)

        main_box = Gtk.Box()
        main_box.set_spacing(5)
        main_box.set_border_width(10)
        main_box.set_orientation(Gtk.Orientation.VERTICAL)

        common.progress_bar = Gtk.ProgressBar()
        common.progress_bar.set_fraction(0.0)
        common.progress_bar.set_text('0')
        common.progress_bar.set_show_text(True)
        main_box.pack_start(common.progress_bar, True, False, 0)
        window.add(main_box)
        window.show_all()

        # This contains a Gtk.ScrolledWindow with Gtk.Grid() inside, filled with ThumbButton(Gtk.Button) instances
        common.preview = Preview()

        main_box.pack_start(common.preview, False, False, 0)

        # We need a horizontal container to display outputs in columns
        displays_box = Gtk.Box()
        displays_box.set_spacing(15)
        displays_box.set_orientation(Gtk.Orientation.HORIZONTAL)

        # Restore saved wallpapers if any
        f_name = "swaybg.json" if common.sway or common.env['wayland'] else "feh.json"
        f_path = os.path.join(common.data_home, f_name)

        if os.path.isfile(f_path):
            restore_from = load_json(f_path)
        else:
            restore_from = None

        # Buttons below represent displays preview
        common.display_boxes_list = []
        for display in common.displays:
            name = display.get('name')
            # Check if we have stored values
            path, thumb = None, None
            if restore_from:
                for item in restore_from:
                    if item["name"] == name:
                        path = item["path"]
                        thumb = item["thumb"]

            # Label format: name (width x height)
            try:
                xrandr_idx = display.get('xrandr-idx')
            except KeyError:
                xrandr_idx = None
            display_box = DisplayBox(name, display.get('width'), display.get('height'), path, thumb, xrandr_idx)
            common.display_boxes_list.append(display_box)
            displays_box.pack_start(display_box, True, False, 0)

        main_box.pack_start(displays_box, False, False, 0)

        # Bottom buttons will also need a horizontal container
        bottom_box = Gtk.Box()
        bottom_box.set_spacing(5)
        bottom_box.set_border_width(5)
        bottom_box.set_orientation(Gtk.Orientation.HORIZONTAL)

        # Button to change sorting order
        sorting_button = SortingButton()
        bottom_box.add(sorting_button)

        # Button to refresh currently selected folder thumbnails
        refresh_button = Gtk.Button()
        refresh_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_refresh.svg'))
        refresh_button.set_image(img)
        refresh_button.set_tooltip_text(common.lang['refresh_folder_preview'])
        bottom_box.add(refresh_button)
        refresh_button.connect_after('clicked', on_refresh_clicked)

        # Button to set the wallpapers folder
        folder_button = Gtk.Button.new_with_label(common.settings.src_path)
        folder_button.set_property("name", "folder-btn")
        folder_button.set_tooltip_text(common.lang['open_another_folder'])
        bottom_box.pack_start(folder_button, True, True, 0)
        folder_button.connect_after('clicked', on_folder_clicked)

        # Label to display details of currently selected picture
        common.selected_picture_label = Gtk.Label()
        common.selected_picture_label.set_property("name", "selected-label")
        common.selected_picture_label.set_text(common.lang['no_picture_selected'])

        bottom_box.pack_start(common.selected_picture_label, True, True, 0)

        # Button to split wallpaper between displays
        if len(common.displays) > 1:
            common.split_button = Gtk.Button()
            common.split_button.set_always_show_image(True)
            img = Gtk.Image()
            img.set_from_file(os.path.join(dir_name, 'images/icon_split.svg'))
            common.split_button.set_image(img)
            bottom_box.add(common.split_button)
            common.split_button.set_sensitive(False)
            common.split_button.set_tooltip_text(common.lang['split_selection_between_displays'])
            common.split_button.connect('clicked', on_split_button)

        # Button to apply selected wallpaper to all displays (connected at the moment or not)
        common.apply_to_all_button = Gtk.Button()
        common.apply_to_all_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_all.svg'))
        common.apply_to_all_button.set_image(img)
        common.apply_to_all_button.connect('clicked', on_apply_to_all_button)
        common.apply_to_all_button.set_sensitive(False)
        common.apply_to_all_button.set_tooltip_text(common.lang['apply_to_all'])
        bottom_box.add(common.apply_to_all_button)

        # Button to apply settings
        names = ''
        for display in common.displays:
            names += '{} '.format(display['name'])

        common.apply_button = Gtk.Button()
        common.apply_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_apply.svg'))
        common.apply_button.set_image(img)
        common.apply_button.connect('clicked', on_apply_button)
        common.apply_button.set_sensitive(False)
        common.apply_button.set_tooltip_text(common.lang['apply_settings'].format(names))
        bottom_box.add(common.apply_button)

        main_box.add(bottom_box)

        h_separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        main_box.add(h_separator)

        # Another horizontal container for the status line + button(s)
        status_box = Gtk.Box()
        status_box.set_spacing(5)
        status_box.set_border_width(5)
        status_box.set_orientation(Gtk.Orientation.HORIZONTAL)

        # Button to call About dialog
        about_button = Gtk.Button()
        about_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_about.svg'))
        about_button.set_image(img)
        about_button.set_tooltip_text(common.lang['about_azote'])
        about_button.connect('clicked', on_about_button)
        status_box.add(about_button)

        # Button to display settings menu
        settings_button = Gtk.Button()
        settings_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_menu.svg'))
        settings_button.set_image(img)
        settings_button.set_tooltip_text(common.lang['preferences'])
        settings_button.connect('clicked', on_settings_button)
        status_box.add(settings_button)

        # Color picker button
        picker_button = Gtk.Button()
        picker_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_picker.svg'))
        picker_button.set_image(img)
        picker_button.set_sensitive(common.picker)
        if common.sway or common.env['wayland']:
            tt = common.lang['screen_color_picker'] if common.picker else common.lang['grim_slurp_required']
        else:
            tt = common.lang['screen_color_picker'] if common.picker else common.lang['maim_slop_required']
        picker_button.set_tooltip_text(tt)
        picker_button.connect('clicked', on_picker_button)
        status_box.add(picker_button)

        # dotfiles button
        dotfiles_button = Gtk.Button()
        dotfiles_button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_config.svg'))
        dotfiles_button.set_image(img)
        active = common.xresources or common.alacritty_config and common.env['yaml']
        if active:
            dotfiles_button.set_tooltip_text(common.lang['dotfiles'])
        else:
            dotfiles_button.set_tooltip_text(common.lang['check_log'].format(common.log_file))
            dotfiles_button.set_sensitive(False)
        dotfiles_button.connect('clicked', on_dotfiles_button)
        status_box.add(dotfiles_button)

        common.status_bar = Gtk.Statusbar()
        common.status_bar.set_property("name", "status-bar")
        common.status_bar.set_halign(Gtk.Align.CENTER)
        status_box.pack_start(common.status_bar, True, True, 0)
        update_status_bar()

        btn = Gtk.Button.new_with_label(common.lang["close"])
        btn.connect("clicked", Gtk.main_quit)
        btn.set_property("valign", Gtk.Align.CENTER)
        status_box.pack_end(btn, False, False, 0)

        main_box.add(status_box)

        window.show_all()
        deselect_all()

        common.progress_bar.hide()

    def handle_keyboard(self, item, event):
        if event.type == Gdk.EventType.KEY_RELEASE and event.keyval == Gdk.KEY_Escape:
            Gtk.main_quit()
        return True


def on_apply_to_all_button(button):
    """
    This will create a single command to set the same wallpaper to all displays, CONNECTED at the time OR NOT.
    Menu for modes needs to differ for swaybg and feh.
    """
    menu = Gtk.Menu()
    if common.sway or common.env['wayland']:
        for mode in common.modes_swaybg:
            item = Gtk.MenuItem.new_with_label(mode)
            item.connect('activate', apply_to_all_swaybg, mode)
            menu.append(item)
        menu.show_all()
        menu.popup_at_widget(button, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH_EAST, None)
    else:
        for mode in common.modes_feh:
            item = Gtk.MenuItem.new_with_label(mode)
            item.connect('activate', apply_to_all_feh, mode)
            menu.append(item)
        menu.show_all()
        menu.popup_at_widget(button, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH_EAST, None)


def on_settings_button(button):
    menu = Gtk.Menu()

    item = Gtk.MenuItem.new_with_label(common.lang['custom_display'])
    item.connect('activate', show_custom_display_dialog)
    menu.append(item)

    item = Gtk.CheckMenuItem.new_with_label(common.lang['color_dictionary'])
    item.set_active(common.settings.color_dictionary)
    item.connect('activate', switch_color_dictionary)
    menu.append(item)

    item = Gtk.CheckMenuItem.new_with_label(common.lang['image_menu_button'])
    item.set_active(common.settings.image_menu_button)
    item.connect('activate', switch_image_menu_button)
    menu.append(item)

    item = Gtk.CheckMenuItem.new_with_label(common.lang['track_file_changes'])
    item.set_active(common.settings.track_files)
    item.connect('activate', switch_tracking_files)
    menu.append(item)

    item = Gtk.CheckMenuItem.new_with_label(common.lang['use_display_names'])
    item.set_active(common.settings.generic_display_names)
    item.connect('activate', switch_generic_display_names)
    menu.append(item)

    menu.show_all()
    menu.popup_at_widget(button, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH_WEST, None)


def pick_color():
    """
    In sway we'll just use grim & slurp to pick a color: it returns accurate values.
    On X11 same should be possible with maim & slop, but it happens to crash.
    In both cases we'll use the same fallback: calculate the dominant colour of a selected region with colorthief.
    This is less accurate, alas.
    :return: tuple (rrr, ggg, bbb)
    """
    color = (255, 255, 255)
    if common.sway or common.env['wayland']:
        try:
            color = hex_to_rgb(subprocess.check_output(
                'grim -g "$(slurp -p)" -t ppm - | magick - -format \'%[pixel:p{0,0}]\' txt:- | awk \'NR==2 {print $3}\'',
                shell=True).decode("utf-8"))
        except:
            try:
                color = get_dominant_from_area()
            except:
                pass
    else:
        try:
            output = subprocess.check_output('maim -st 0 | magick - -resize 1x1! -format \'%[pixel:p{0,0}]\' info:-',
                                             shell=True).decode("utf-8")
            values = output[6:-1].split(",")
            color = (int(values[0]), int(values[1]), int(values[2]))
        except:
            try:
                color = get_dominant_from_area()
            except:
                pass
    return color


def get_dominant_from_area():
    """
    Saves selected area with `grim -g "$(slurp)" common.tmp_dir/azote/temp/area.png`,
    then calculates the dominant color with the colorthief module.
    :return: tuple (r, g, b) or (255, 255, 255) if nothing selected
    """
    dominant = (255, 255, 255)
    if common.sway or common.env['wayland']:
        cmd = 'grim -g "$(slurp)" {}'.format(os.path.join(common.tmp_dir, 'area.png'))
    else:
        cmd = 'maim -s {}'.format(os.path.join(common.tmp_dir, 'area.png'))

    res = subprocess.run(cmd, shell=True, stdout=subprocess.DEVNULL).returncode
    if res == 0:
        color_thief = ColorThief(os.path.join(common.tmp_dir, 'area.png'))
        try:
            dominant = color_thief.get_color(quality=common.settings.palette_quality)
        except:
            pass

    return dominant


def on_picker_button(button):
    if common.picker_window:
        common.picker_window.close()
    color = pick_color()
    common.picker_window = ColorPickerDialog(color)


def on_dotfiles_button(button):
    if common.xresources or common.alacritty_config:
        menu = Gtk.Menu()
        if common.xresources:
            item = Gtk.MenuItem.new_with_label(common.xresources)
            item.connect('activate', open_dotfile, 'xresources')
            menu.append(item)

        if common.env['yaml'] and common.alacritty_config:
            item = Gtk.MenuItem.new_with_label(common.alacritty_config)
            item.connect('activate', open_dotfile, 'alacritty')
            menu.append(item)
        menu.show_all()
        menu.popup_at_widget(button, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH_WEST, None)


def open_dotfile(widget, which):
    if common.dotfile_window:
        common.dotfile_window.close()

    if which == 'alacritty' and common.alacritty_config:
        common.dotfile_window = Alacritty()

    elif which == 'xresources' and common.xresources:
        common.dotfile_window = Xresources()


def show_custom_display_dialog(item):
    cdd = CustomDisplayDialog()


def switch_color_dictionary(item):
    if item.get_active():
        common.settings.color_dictionary = True
        common.settings.save()
    else:
        common.settings.color_dictionary = False
        common.settings.save()


def switch_image_menu_button(item):
    if item.get_active():
        common.settings.image_menu_button = True
        common.settings.save()
    else:
        common.settings.image_menu_button = False
        common.settings.save()


def switch_tracking_files(item):
    if item.get_active():
        common.settings.track_files = True
        common.settings.save()
        GLib.timeout_add_seconds(common.settings.tracking_interval_seconds, track_changes)
    else:
        common.settings.track_files = False
        common.settings.save()
    if common.indicator:
        common.indicator.switch_indication(item)


def switch_generic_display_names(item):
    if item.get_active():
        common.settings.generic_display_names = True
        common.settings.save()
    else:
        common.settings.generic_display_names = False
        common.settings.save()


class ColorPaletteDialog(Gtk.Window):
    def __init__(self, thumb_file, filename, palette):
        super().__init__()

        self.image = Gtk.Image.new_from_file(thumb_file)
        self.label = Gtk.Label()
        self.label.set_text(filename)
        self.label.set_property('name', 'image-label')

        self.set_title(filename)
        self.set_role("toolbox")
        self.set_resizable(False)
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        # self.set_transient_for(common.main_window)
        self.set_position(Gtk.WindowPosition.MOUSE)
        self.set_keep_above(True)
        self.all_buttons = []

        try:
            self.copy_as = common.settings.copy_as
        except AttributeError:
            self.copy_as = '#rgb'

        self.vbox = Gtk.VBox()
        self.vbox.set_spacing(5)
        self.vbox.set_border_width(15)

        self.vbox.pack_start(self.image, True, True, 15)
        self.vbox.add(self.label)

        self.hbox = Gtk.HBox()
        self.hbox.set_spacing(5)
        self.hbox.set_border_width(5)

        for i in range(len(palette)):
            color = palette[i]
            hex_color = rgb_to_hex(color)

            pixbuf = create_pixbuf((common.settings.color_icon_w, common.settings.color_icon_h), color)
            gtk_image = Gtk.Image.new_from_pixbuf(pixbuf)

            button = Gtk.Button.new_with_label(hex_color)
            button.set_always_show_image(True)
            button.set_image(gtk_image)
            button.set_image_position(2)  # TOP
            if common.settings.color_dictionary:
                exact, closest = common.color_names.get_colour_name(hex_color)
                if exact:
                    name = common.lang['exact'].format(exact)
                else:
                    name = common.lang['closest'].format(closest)
                button.set_tooltip_text('{}'.format(name))
            else:
                button.set_tooltip_text(common.lang['copy'])

            button.connect_after('clicked', self.to_clipboard)
            self.all_buttons.append(button)

            self.hbox.pack_start(button, False, False, 0)

            if (i + 1) % 6 == 0:
                self.vbox.add(self.hbox)
                self.hbox = Gtk.HBox()
                self.hbox.set_spacing(5)
                self.hbox.set_border_width(5)

        button1 = Gtk.RadioButton(label='#rgb')
        button1.connect('toggled', self.rgb_toggled)

        button2 = Gtk.RadioButton.new_with_label_from_widget(button1, 'r, g, b')
        button2.set_label('r, g, b')
        button2.connect('toggled', self.rgb_toggled)

        for button in [button1, button2]:
            if button.get_label() == self.copy_as:
                button.set_active(True)

        label = Gtk.Label()
        label.set_text(common.lang['copy_as'])

        hbox = Gtk.HBox()
        hbox.set_spacing(5)
        hbox.set_border_width(5)

        hbox.pack_start(label, True, False, 0)

        hbox.pack_start(button1, True, False, 0)
        hbox.pack_start(button2, True, False, 0)

        self.vbox.pack_start(hbox, True, True, 0)

        hbox = Gtk.HBox()
        hbox.set_spacing(5)
        hbox.set_border_width(5)

        self.clipboard_preview = ClipboardPreview()
        hbox.pack_start(self.clipboard_preview, False, False, 0)

        self.clipboard_label = Gtk.Label()
        self.clipboard_label.set_text(common.lang['clipboard_empty'])
        hbox.pack_start(self.clipboard_label, True, True, 0)

        button = Gtk.Button.new_with_label(common.lang['close'])
        button.connect_after('clicked', self.close_window)
        hbox.pack_start(button, False, False, 0)

        self.vbox.add(hbox)

        self.add(self.vbox)
        self.show_all()

    def rgb_toggled(self, button):
        state = 'on' if button.get_active() else 'off'
        if state == 'on':
            common.settings.copy_as = button.get_label()
            common.settings.save()

    def close_window(self, widget):
        self.close()

    def show(self):
        self.show_all()

    def to_clipboard(self, widget):
        # clear selection
        for button in self.all_buttons:
            button.set_property("name", "color-btn")
        # mark selected
        widget.set_property("name", "color-btn-selected")

        common.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        if common.settings.copy_as == 'r, g, b':
            t = hex_to_rgb(widget.get_label())
            content = '{}, {}, {}'.format(t[0], t[1], t[2])
        else:
            content = widget.get_label()

        common.clipboard.set_text(content, -1)
        common.clipboard_text = widget.get_label()

        label = '{}'.format(content)
        self.clipboard_preview.update(widget.get_label())
        self.clipboard_label.set_text(label)


class ClipboardPreview(Gtk.ColorButton):
    def __init__(self):
        super().__init__()
        rgba = rgb_to_rgba((255, 255, 255))

        color = Gdk.RGBA()
        color.red = rgba[0]
        color.green = rgba[1]
        color.blue = rgba[2]
        color.alpha = rgba[3]
        self.set_rgba(color)
        self.connect("color-set", self.to_clipboard)

    def update(self, color):
        rgb = hex_to_rgb(color)
        rgba = rgb_to_rgba(rgb)

        color = Gdk.RGBA()
        color.red = rgba[0]
        color.green = rgba[1]
        color.blue = rgba[2]
        color.alpha = rgba[3]
        self.set_rgba(color)

    def to_clipboard(self, widget):
        common.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        rgba = self.get_rgba()
        if common.settings.copy_as == 'r, g, b':
            hex = rgba_to_hex(rgba)
            t = hex_to_rgb(hex)
            content = '{}, {}, {}'.format(t[0], t[1], t[2])
        else:
            content = rgba_to_hex(rgba)
            hex = content

        common.clipboard.set_text(content, -1)
        common.clipboard_text = hex
        label = '{}'.format(content)
        common.cpd.clipboard_label.set_text(label)


class ColorPickerDialog(Gtk.Window):
    def __init__(self, color):
        super().__init__()

        if not color:
            color = (255, 255, 255)

        self.set_title(common.lang['screen_color_picker'])
        self.set_role("toolbox")
        self.set_resizable(False)
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        # self.set_transient_for(common.main_window)
        self.set_position(Gtk.WindowPosition.MOUSE)
        self.set_keep_above(True)

        try:
            self.copy_as = common.settings.copy_as
        except AttributeError:
            self.copy_as = '#rgb'

        self.vbox = Gtk.VBox()
        self.vbox.set_spacing(5)
        self.vbox.set_border_width(5)

        rgba = rgb_to_rgba(color)

        self.color_button = Gtk.ColorButton()
        rgba_color = Gdk.RGBA()
        rgba_color.red = rgba[0]
        rgba_color.green = rgba[1]
        rgba_color.blue = rgba[2]
        rgba_color.alpha = 1.0
        self.color_button.set_rgba(rgba_color)
        self.color_button.connect("color-set", self.on_color_chosen, self.color_button)
        self.color_button.set_tooltip_text(common.lang['background_color'])
        self.vbox.pack_start(self.color_button, False, False, 0)

        self.label = Gtk.Label()
        self.label.set_text(rgb_to_hex(color))
        self.vbox.add(self.label)

        if common.settings.color_dictionary:
            self.closest_label = Gtk.Label()
            self.closest_label.set_property('name', 'closest')
            exact, closest = common.color_names.get_colour_name(rgb_to_hex(color))
            if exact:
                name = common.lang['exact'].format(exact)
            else:
                name = common.lang['closest'].format(closest)
            self.closest_label.set_text(name)
            self.vbox.add(self.closest_label)

        button1 = Gtk.RadioButton(label='#rgb')
        button1.connect('toggled', self.rgb_toggled)

        button2 = Gtk.RadioButton.new_with_label_from_widget(button1, 'r, g, b')
        button2.set_label('r, g, b')
        button2.connect('toggled', self.rgb_toggled)

        for button in [button1, button2]:
            if button.get_label() == self.copy_as:
                button.set_active(True)

        hbox = Gtk.HBox()
        hbox.set_spacing(5)
        hbox.set_border_width(5)

        hbox.pack_start(button1, True, False, 0)
        hbox.pack_start(button2, True, False, 0)

        self.vbox.add(hbox)

        hbox = Gtk.HBox()
        hbox.set_spacing(5)
        hbox.set_border_width(5)

        button = Gtk.Button.new_with_label(common.lang['copy'])
        button.connect_after('clicked', self.to_clipboard)
        hbox.pack_start(button, True, False, 0)

        button = Gtk.Button()
        button.set_always_show_image(True)
        img = Gtk.Image()
        img.set_from_file(os.path.join(dir_name, 'images/icon_picker.svg'))
        button.set_image(img)
        button.connect_after('clicked', self.pick_new_color)
        hbox.pack_start(button, True, False, 0)

        button = Gtk.Button.new_with_label(common.lang['close'])
        button.connect_after('clicked', self.close_window)
        hbox.pack_start(button, True, False, 0)

        self.vbox.add(hbox)

        self.add(self.vbox)
        self.show_all()

    def on_color_chosen(self, user_data, button):
        self.label.set_text(rgba_to_hex(button.get_rgba()))
        exact, closest = common.color_names.get_colour_name(self.label.get_text())
        if exact:
            name = common.lang['exact'].format(exact)
        else:
            name = common.lang['closest'].format(closest)
        self.closest_label.set_text(name)

    def rgb_toggled(self, button):
        state = 'on' if button.get_active() else 'off'
        if state == 'on':
            common.settings.copy_as = button.get_label()
            common.settings.save()

    def close_window(self, widget):
        self.close()

    def to_clipboard(self, widget):
        common.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        if common.settings.copy_as == 'r, g, b':
            t = hex_to_rgb(self.label.get_text())
            content = '{}, {}, {}'.format(t[0], t[1], t[2])
        else:
            content = self.label.get_text()

        common.clipboard.set_text(content, -1)
        common.clipboard_text = self.label.get_text()

    def pick_new_color(self, button):
        color = pick_color()
        if not color:
            color = (255, 255, 255)

        rgba = rgb_to_rgba(color)
        rgba_color = Gdk.RGBA()
        rgba_color.red = rgba[0]
        rgba_color.green = rgba[1]
        rgba_color.blue = rgba[2]
        rgba_color.alpha = 1.0
        self.color_button.set_rgba(rgba_color)

        self.label.set_text(rgb_to_hex(color))

        if common.settings.color_dictionary:
            exact, closest = common.color_names.get_colour_name(rgb_to_hex(color))
            if exact:
                name = common.lang['exact'].format(exact)
            else:
                name = common.lang['closest'].format(closest)
            self.closest_label.set_text(name)
            self.vbox.add(self.closest_label)


class CustomDisplayDialog(Gtk.Window):
    def __init__(self):
        super().__init__()

        self.properties = common.settings.custom_display

        self.set_title("Azote custom display")
        self.set_role("toolbox")
        self.set_resizable(False)
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.set_decorated(False)
        # self.set_transient_for(common.main_window)
        self.set_position(Gtk.WindowPosition.MOUSE)
        self.set_keep_above(True)

        self.name_label = Gtk.Label()
        self.name_label.set_width_chars(12)
        self.name_label.set_text(common.lang['name'])

        self.name_entry = Gtk.Entry()
        if self.properties and self.properties[0]:
            self.name_entry.set_text(self.properties[0])
        self.name_entry.connect('changed', self.validate_entries)

        self.width_label = Gtk.Label()
        self.width_label.set_width_chars(12)
        self.width_label.set_text(common.lang['width'])

        self.button_ok = Gtk.Button.new_with_label(common.lang['ok'])
        self.button_ok.set_sensitive(False)

        self.width_entry = NumberEntry()
        if self.properties:
            self.width_entry.set_text(self.properties[1])
        self.width_entry.connect('changed', self.validate_entries)

        self.height_label = Gtk.Label()
        self.height_label.set_width_chars(12)
        self.height_label.set_text(common.lang['height'])

        self.height_entry = NumberEntry()
        if self.properties:
            self.height_entry.set_text(self.properties[2])
        self.height_entry.connect('changed', self.validate_entries)

        self.button_cancel = Gtk.Button.new_with_label(common.lang['cancel'])
        self.button_cancel.connect("clicked", self.dialog_cancel, self)

        self.button_clear = Gtk.Button.new_with_label(common.lang['delete'])
        self.button_clear.connect("clicked", self.dialog_clear, self)

        self.vbox = Gtk.VBox()
        self.vbox.set_spacing(5)
        self.vbox.set_border_width(5)

        self.hbox0 = Gtk.HBox()
        self.hbox0.pack_start(self.name_label, True, True, 0)
        self.hbox0.add(self.name_entry)
        self.vbox.add(self.hbox0)

        self.hbox1 = Gtk.HBox()
        self.hbox1.pack_start(self.width_label, True, True, 0)
        self.hbox1.add(self.width_entry)
        self.vbox.add(self.hbox1)

        self.hbox2 = Gtk.HBox()
        self.hbox2.pack_start(self.height_label, True, True, 0)
        self.hbox2.add(self.height_entry)
        self.vbox.add(self.hbox2)

        self.hbox3 = Gtk.HBox()
        self.hbox3.pack_start(self.button_ok, True, True, 0)
        self.hbox3.pack_start(self.button_cancel, True, True, 5)
        self.hbox3.pack_start(self.button_clear, True, True, 0)
        self.vbox.pack_start(self.hbox3, True, True, 0)

        self.add(self.vbox)
        self.button_ok.connect("clicked", self.dialog_ok)
        self.show_all()

    def validate_entries(self, widget):
        self.button_ok.set_sensitive(self.width_entry.get_text() and self.height_entry.get_text())

    def dialog_ok(self, widget, callback_data=None):
        self.properties = [self.name_entry.get_text(), self.width_entry.get_text(), self.height_entry.get_text()]
        if not self.properties[0]:
            self.properties[0] = 'Custom'
        common.settings.custom_display = self.properties
        common.settings.save()
        self.close()

    def dialog_cancel(self, widget, callback_data=None):
        self.close()

    def dialog_clear(self, widget, callback_data=None):
        common.settings.custom_display = None
        common.settings.save()
        self.close()


class NumberEntry(Gtk.Entry):
    """
    https://stackoverflow.com/a/2727085/4040598
    """

    def __init__(self):
        Gtk.Entry.__init__(self)
        self.connect('changed', self.on_changed)

    def on_changed(self, *args):
        text = self.get_text().strip()
        self.set_text(''.join([i for i in text if i in '0123456789']))


def dialog_cancel(widget, window, callback_data=None):
    window.close()


def on_thumb_double_click(button):
    """
    As the function above, but mode 'fill' will always be used
    """
    if common.sway or common.env['wayland']:
        apply_to_all_swaybg(button, 'fill')
    else:
        apply_to_all_feh(button, 'fill')


def apply_to_all_swaybg(item, mode):
    # Firstly we need to set the selected image thumbnail to all previews currently visible
    for box in common.display_boxes_list:
        box.img.set_from_file(common.selected_wallpaper.thumb_file)
        box.wallpaper_path = common.selected_wallpaper.source_path
        box.thumbnail_path = common.selected_wallpaper.thumb_file

    common.apply_button.set_sensitive(True)


def apply_to_all_feh(item, mode):
    # Firstly we need to set the selected image thumbnail to all previews currently visible
    for box in common.display_boxes_list:
        box.img.set_from_file(common.selected_wallpaper.thumb_file)
        box.wallpaper_path = common.selected_wallpaper.source_path
        box.thumbnail_path = common.selected_wallpaper.thumb_file

    common.apply_button.set_sensitive(True)


def print_help():
    print('\nAzote wallpaper manager version {}\n'.format(__version__))
    print('[-h] | [--help]\t\t\t print Help')
    print('[-l] | [--lang] <ln_LN> \t force a Locale (de_DE, en_US, fr_FR, pl_PL)')
    print('[-c] | [--clear]\t\t Clear unused thumbnails')
    print('[-a] | [--clear-all]\t\t clear All thumbnails\n')
    print('[-v] | [--version]\t\t display Version information\n')


def track_changes():
    if common.preview and common.settings.src_path:
        files_dict = dict([(f, None) for f in os.listdir(common.settings.src_path)])
        if not files_dict == common.preview.files_dict:
            common.preview.refresh()
    return common.settings.track_files


class Indicator(object):
    def __init__(self):
        self.ind = AppIndicator3.Indicator.new('azote_status_icon', '',
                                               AppIndicator3.IndicatorCategory.APPLICATION_STATUS)
        self.ind.set_icon_full('/usr/share/azote/indicator_active.png', 'Tracking off')
        self.ind.set_attention_icon_full('/usr/share/azote/indicator_attention.png', 'Tracking on')

        if common.settings.track_files:
            self.ind.set_status(AppIndicator3.IndicatorStatus.ATTENTION)
            if common.sway or common.env['wayland']:
                self.ind.set_icon_full('/usr/share/azote/indicator_attention.png', 'Tracking on')
        else:
            self.ind.set_status(AppIndicator3.IndicatorStatus.ACTIVE)
            if common.sway or common.env['wayland']:
                self.ind.set_icon_full('/usr/share/azote/indicator_active.png', 'Tracking off')

        self.ind.set_menu(self.menu())

    def menu(self):
        menu = Gtk.Menu()

        item = Gtk.MenuItem.new_with_label(common.lang['clear_unused_thumbnails'])
        item.connect('activate', self.clear_unused)
        menu.append(item)

        item = Gtk.MenuItem.new_with_label(common.lang['about_azote'])
        item.connect('activate', on_about_button)
        menu.append(item)

        item = Gtk.SeparatorMenuItem()
        menu.append(item)

        item = Gtk.MenuItem.new_with_label(common.lang['exit'])
        item.connect('activate', destroy)
        menu.append(item)

        menu.show_all()
        return menu

    def clear_unused(self, item):
        clear_thumbnails()
        common.preview.refresh()

    def switch_indication(self, item):
        if item.get_active():
            self.ind.set_status(AppIndicator3.IndicatorStatus.ATTENTION)
            if common.sway or common.env['wayland']:
                self.ind.set_icon_full('/usr/share/azote/indicator_attention.png', 'Tracking on')
        else:
            if common.sway or common.env['wayland']:
                self.ind.set_icon_full('/usr/share/azote/indicator_active.png', 'Tracking off')
            self.ind.set_status(AppIndicator3.IndicatorStatus.ACTIVE)


def main():
    lang_from_args = None
    clear_thumbs, clear_all = False, False
    common.color_names = WikiColours()
    for i in range(1, len(sys.argv)):
        if sys.argv[i].upper() == '-H' or sys.argv[i].upper() == '--HELP':
            print_help()
            exit(0)

        if sys.argv[i].upper() == '-L' or sys.argv[i].upper() == '--LANG':
            try:
                lang_from_args = sys.argv[i + 1]
            except:
                pass

        if sys.argv[i].upper() == '-C' or sys.argv[i].upper() == '--CLEAR':
            clear_thumbs = True

        if sys.argv[i].upper() == '-A' or sys.argv[i].upper() == '--CLEAR-ALL':
            clear_thumbs, clear_all = True, True

        if sys.argv[i].upper() == '-V' or sys.argv[i].upper() == '--VERSION':
            print("Azote version {}".format(__version__))
            exit(0)

    screen = Gdk.Screen.get_default()
    provider = Gtk.CssProvider()
    style_context = Gtk.StyleContext()
    style_context.add_provider_for_screen(
        screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )
    css = b"""
            button#thumb-btn {
                font-weight: normal;
                font-size: 11px;
            }
            button#folder-btn {
                font-size: 12px;
            }
            button#thumb-btn-selected {
                font-weight: bold;
                font-size: 12px;
                border-top: 1px solid #ccc;
                border-left: 1px solid #ccc;
                border-bottom: 1px solid #333;
                border-right: 1px solid #333;
            }
            button#color-btn {
                font-weight: normal;
            }
            button#color-btn-selected {
                font-weight: bold;
                border-top: 1px solid #ccc;
                border-left: 1px solid #ccc;
                border-bottom: 1px solid #333;
                border-right: 1px solid #333;
            }

            button#display-btn {
                font-weight: normal;
                font-size: 12px;
            }
            button#display-btn-selected {
                font-weight: bold;
                font-size: 12px;
            }
            label#image-label {
                font-size: 12px;
            }
            statusbar#status-bar {
                font-size: 12px;
            }
            label#selected-label {
                border-top: 1px solid #ccc;
                border-left: 1px solid #ccc;
                border-bottom: 1px solid #333;
                border-right: 1px solid #333;
                font-size: 12px;
            }
            label#dotfiles {
                font-size: 13px;
                margin-top: 0px;
                margin-bottom: 0px;
                margin-right: 10px;
            }
            label#dotfiles-header {
                font-size: 14px;
                font-weight: bold;
            }
            label#closest {
                font-size: 12px;
            }
            button#dotfiles-button {
                padding: 1px;
                margin: 0px;
                font-size: 12px;
            }
            textview#preview {
                font-size: 13px;
            }
            """
    provider.load_from_data(css)

    set_env(__version__, lang_from_args=lang_from_args)  # detect displays, check installed modules, set paths and stuff
    if clear_thumbs:
        clear_thumbnails(clear_all)
        exit()

    common.cols = len(common.displays) if len(common.displays) > common.settings.columns else common.settings.columns

    if common.settings.track_files:
        GLib.timeout_add_seconds(common.settings.tracking_interval_seconds, track_changes)
    if common.env['app_indicator']:
        common.indicator = Indicator()

    # We want Azote to take all the possible screen height. Since Gdk.Screen.height is deprecated, we need to measure
    # the current screen height in another way. `w` is a temporary window.
    # If on sway, we've already detected the screen height in tools/check_displays() and stored it in common.screen_h
    if not common.screen_h:  # neither sway, nor Hyprland
        w = TransparentWindow()
        if common.sway or common.env['wm'] == "i3":
            w.fullscreen()  # .maximize() doesn't work as expected on sway
        else:
            w.maximize()
        w.present()

        GLib.timeout_add(common.settings.screen_measurement_delay, check_height_and_start, w)
    else:
        if os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
            app = GUI(common.screen_h)  # Hyprland
        else:
            app = GUI(int(common.screen_h * 0.95))  # sway

    Gtk.main()


if __name__ == "__main__":
    sys.exit(main())
